package GW::Base;
##############################################
#	Base class for GateWay protocols
#	Copyright: Slavej Karadjov -2003
##############################################
use strict;


sub new {
	my $proto = shift;
	my $class = ref($proto) || $proto;
	my $self = {
			nHostUserID => shift,
			sRemoteUserID => shift,
			sImProtocol => shift,
			sCharset => "cp1251",
			PROTOCOL => "Base",
			CHECK_CONNECT_INTERVAL => 10, # in seconds
			bUserIsConnected => 0,
			nLastConnectedCheck => 0,
			_aOutQueue => [],
			_aInQueue => [],
			_hGroups => {},
			_hUsers => {},
		    };
	bless ($self, $class);
}

sub getCharset {
	my $self=shift;
	return $self->{sCharset};
}

sub isUserConnected {
	my $self = shift;
	$self->_debug("isUserConnected(): -- NOT IMPLEMENTED --");
	return $self->{bUserIsConnected};
}
sub hasOutgoingEvents {
	my $self = shift;
	$self->getOutgoingEvents();
	return @{$self->{_aOutQueue}};
}

sub getOutgoingEvents {
	my $self = shift();
	$self->_debug("getOutgoingEvents(): -- NOT IMPLEMENTED --");
}

sub getOutgoingEventsFromQueue {
	my $self = shift;
	my @aResult = @{$self->{_aOutQueue}};
	@{$self->{_aOutQueue}} = ();
	return @aResult;
}


sub hasIncomingEvents{
	my $self = shift;
	$self->_debug("hasIncomingEvents(): -- NOT IMPLEMENTED --");
	return 0;
}

sub addEventsToQueue {
	my $self = shift();
	my @aEvent = @_;
	
	foreach(@aEvent) {		
		$self->_writeEvent($_);
	}
}

sub keepAlive { # do we need to keep open connection to the IM proxy server
	my $self = shift;
	$self->_debug("keepAlive(): -- NOT IMPLEMENTED --");
	return 0;
}

sub sendContactList {
	my ($self, $sRawContactList) = @_;
	my ($sEncoding, $sGroups, $sUsers) = ();
	my ($i, $max) = ();
	
	if($sRawContactList =~/^(.*?);\[(.*?)\];\[(.*?)\]$/) {
		$sEncoding = $1;	
		$sGroups = $2;
		$sUsers =$3;
	}
	
	# process groups	
	my @pair = split(/\|/,$sGroups);
	foreach(@pair) {
		my ($sGroupName, $sGroupID)= split(/,/);
		${$self->{_hGroups}}{$sGroupID}->{'NAME'} = $sGroupName;
	}
	
	# process users	
	@pair = split(/\|/,$sUsers);
	foreach(@pair) {	
		if($_=~/^(.*?),(.*?),\{(.*?)\}$/) {		
			my $sUserID = $1;
			my $sUserName = $2;
			my $sUserGroupList = $3;
			${$self->{_hUsers}}{$sUserID}->{'NAME'} = $sUserName;
			my @aGroupList = split(/,/,$sUserGroupList);		
			foreach(@aGroupList) {			
				push(@{${$self->{_hGroups}}{$_}->{'USER_LIST'}}, $sUserID);
			}	
		}	
	}

#ToDo:: Send user/group data to BGDATE IM client	 
	print "\n-----------------------------\n";
	foreach(keys(%{$self->{_hGroups}})) {
		print "--".${$self->{_hGroups}}{$_}->{'NAME'}."--\n";
		foreach my $tUserID (@{ ${$self->{_hGroups}}{$_}->{'USER_LIST'} }) {
			print $_.":".${$self->{_hUsers}}{$tUserID}->{'NAME'}."\n";
		}
	}
	print "\n-----------------------------\n";	
	
	# update user list
	foreach(keys(%{$self->{_hUsers}})) {
		$self->_sendEvent('SYS', 'ADD_USR', $self->{nHostUserID}, $self->{sImProtocol}."_".$_, ${$self->{_hUsers}}{$_}->{'NAME'}, time());
	}

	# update group list
	foreach(keys(%{$self->{_hGroups}})) {
		print "--".${$self->{_hGroups}}{$_}->{'NAME'}."--\n";

		for($i=0, $max=@{ ${$self->{_hGroups}}{$_}->{'USER_LIST'}}; $i<$max; $i++)  {
			${ ${$self->{_hGroups}}{$_}->{'USER_LIST'}}[$i]=$self->{sImProtocol}."_".${ ${$self->{_hGroups}}{$_}->{'USER_LIST'}}[$i]
		}

		$self->_sendEvent('SYS', 'ADD_GROUP', $self->{nHostUserID},-1,
				  "id=".$self->{sImProtocol}."_".$_."&name=".$self->_urlencode(${$self->{_hGroups}}{$_}->{'NAME'}).
				  "&proto=".$self->{sImProtocol}."&user_list=".$self->_urlencode(join(",",@{ ${$self->{_hGroups}}{$_}->{'USER_LIST'}}) ).
				  "&visible=Y&deletable=Y", time());
	}

	# and finally
	#$self->_sendEvent('SYS', 'JS_PRINT', $self->{nHostUserID},-1, "window.top.hProtocol[\"$self->{sImProtocol}\"][\"LOADED\"]=true;\n", time());

	$self->_sendEvent('PRT', 'CONNECTED', $self->{nHostUserID},$self->{sImProtocol}, 1, time());
	$self->_sendEvent('PRT', 'STAT', $self->{nHostUserID},$self->{sImProtocol}, "ONL", time());

}

sub brokenConnection {
	my $self=shift;
	$self->_debug("brokenConnection(): -- NOT IMPLEMENTED --");
}

sub shutdown {
	my $self = shift();
	$self->_debug("shutdown():-- NOT IMPLEMENTED --");
	#            IM Proxy .
}

#:: private functions #

sub _debug {
	my $self = shift();
	my $sText  = shift();	
	print "GW::".$self->{PROTOCOL}." module [ ".$self->{nHostUserID}." - ".(scalar localtime())."] : $sText\n";
}

sub _sendEvent {
	my $self = shift();
	my ($sType, $sCMD, $nToID, $nFromID, $hAttr, $dDate, $dExpire, $bUpdateInbox) = @_;
	$bUpdateInbox=1 if (!defined($bUpdateInbox));
	$dExpire = 0 if (!defined($dExpire));
	$hAttr ='' if (!defined($hAttr));
	
	# [don't send messages to myself] #
	if($nFromID eq $nToID) { 
		return 0;
	}
	
	$self->_debug("_sendEvent(): $sType, $sCMD, $nToID, $nFromID, $hAttr, $dDate, $dExpire, $bUpdateInbox");
	$self->_debug("_sendEvent(): -- NOT IMPLEMENTED --");
	return 1;

}

sub _writeEvent {
	my $self = shift();
	my $sEvent = shift();
	my $sEventHeader = substr($sEvent,0,index($sEvent,"\n")-1);
	my ($enc, $from, $type, $cmd, $time) = split(/\|/,$sEventHeader);
	my $sEventAttr = substr($sEvent,index($sEvent,"\n")+1); 
	$sEventAttr =~s/\n/ /g; # clean newlines
	$sEventAttr =~s/\r//g; # clean carriage returns	
	
#ToDo:: convert from $enc to $self->getCharset();
	
	if($type eq "GW") {
		# This type of commands are only for the GW module
		if($cmd eq "ADD_NEW_GROUP") {
			${$self->{_hGroups}}{$from}->{'NAME'} = $sEventAttr;
			@{ ${$self->{_hGroups}}{$sEventAttr}->{'USER_LIST'}} = ();
			
			$self->_sendEvent('SYS', 'ADD_GROUP', $self->{nHostUserID},-1,
				  "id=".$self->{sImProtocol}."_".$from."&name=".$self->_urlencode(${$self->{_hGroups}}{$from}->{'NAME'}).
				  "&proto=".$self->{sImProtocol}."&user_list=&visible=Y&deletable=Y", time());
		}
		elsif($cmd eq "ADD_USR_TO_GROUP") {
			push(@{ ${$self->{_hGroups}}{$sEventAttr}->{'USER_LIST'}}, $self->{sImProtocol}."_".$from);
			
			$self->_sendEvent('SYS', 'ADD_GROUP', $self->{nHostUserID},-1,
				  "id=".$self->{sImProtocol}."_".$sEventAttr."&name=".$self->_urlencode(${$self->{_hGroups}}{$sEventAttr}->{'NAME'}).
				  "&proto=".$self->{sImProtocol}."&user_list=".$self->_urlencode(join(",",@{ ${$self->{_hGroups}}{$sEventAttr}->{'USER_LIST'}}) ).
				  "&visible=Y&deletable=Y", time());			
		}
		return 0;
	}

	if($self->{sImProtocol}) {
		$from = $self->{sImProtocol}."_".$from;
	}	
	if($cmd eq "NEW_MSG") {
		if($sEventAttr eq "") {
			$self->_debug("_writeEvent($sEventAttr): Empty message?! ");
			return 0;
		}
		$sEventAttr = $self->_reformatMessage($sEventAttr);
	}	
	elsif(($type eq "SYS") && ($cmd eq "NEW_EMAIL")) {
		# $tData[TYPE] -- PopUp Type - ALERT, CLOSE_AFTER, ...
		# $tData[CONTENT] - alert description
		# $tData[TITLE] - alert description
		# $tData[CLOSE] = 0, 1 -- close
		# $tData[TIMEOUT] = timeout in secs
		# $tData[WIDTH] - width
		# $tData[HEIGHT] - width
		
		if($from eq  $self->{sImProtocol}."_-1") {
			$sEventAttr = " $sEventAttr    $self->{sImProtocol} .";
			$sEventAttr = "TYPE=ALERT&CONTENT=".$self->_urlencode($sEventAttr)."&CL=0";			
		}
		else {
			$sEventAttr = "    $self->{sImProtocol}   $from.<br>   : $sEventAttr";
			$sEventAttr = "TYPE=ALERT&CONTENT=".$self->_urlencode($sEventAttr)."&CL=0";			
		}
		$type = "SYS";
		$cmd = "POP_UP";
		$from = -1;
	}
	$self->_sendEvent($type, $cmd, $self->{nHostUserID}, $from, $sEventAttr, time);
}

sub _urlencode {
    my $self = shift();
    my $ask = shift();

    my @a2 = unpack("C*", $ask);
    my $s2 = "";
    while (@a2) {
    	$s2 .= sprintf "%%%X", shift @a2;
    }
    return $s2;
}

sub _reformatMessage {
	my ($self, $sText) = @_;
	$sText=~s/\0$//; # remove ending nulls
	$sText=~s/</&lt;/g;
	$sText=~s/>/&gt;/g;
	$sText=~s/'/&apos;/g;
	$sText=~s/"/&quot;/g;
	return $sText;
}

sub END {
	#ToDo:: send information to client that we are exiting if the connection is active
}

1;
