package GW::LINUXBG;
use base 'GW::Base';
##############################################
#	GateWay for Linux-BG
#	Copyright: Slavej Karadjov - 2003
##############################################
use strict;
use DBI;

my $dbh = undef;

# BEGIN {
# 	eval("use Encode qw(from_to);");
# 	if(@!) {
#  		sub from_to($$$;$) {
# 			return @_;
# 		}
# 	}
# }

sub new {
	my $proto = shift;
	my $class = ref($proto) || $proto;
	my $self = $class->SUPER::new(@_);
	$self->{PROTOCOL} = "LINUXBG";	
	
	my %CONFIG = $self->_loadConfig("_FULL_PATH_TO_/conf/main.conf");		

	$self->_debug($self->{nHostUserID}." is connected");
	$dbh = DBI->connect("DBI:mysql:".$CONFIG{'DB_DATABASE'},$CONFIG{'DB_USER'},$CONFIG{'DB_PASS'}) || die $DBI::errstr;

	bless ($self, $class);
	return $self;
}

sub isUserConnected {
	my $self = shift;
	my $bConnected = 0;

	if(time() < ($self->{nLastConnectedCheck}+$self->{CHECK_CONNECT_INTERVAL})) {
		# don't check too often
		return $self->{bUserIsConnected};
	}

	my $sth1 = $dbh->prepare("SELECT ID FROM IM_SESS WHERE usr='".$self->{nHostUserID}."' ");
	$sth1->execute();
	if($sth1->rows()) {
		$bConnected =1;
	}
	$sth1->finish();

	$self->{nLastConnectedCheck} = time();
	$self->{bUserIsConnected} = $bConnected;
	return $bConnected;
}

sub getOutgoingEvents {
	my $self = shift();
	my @aEventsToDelete = ();
	# fetch outgoing events from DB
	my $sth1 = $dbh->prepare("SELECT ID, TYPE, CMD, TO_USR, ATTR, UNIX_TIMESTAMP(DATE)  FROM IM_EVENT WHERE FROM_USR='".$self->{nHostUserID}."' AND LEFT(TO_USR,3)='".$self->{sImProtocol}."' AND NEW='Y' ");
	$sth1->execute();
	while(my($nEventID, $sType, $sCMD, $sToUser, $sAttr, $sDate) = $sth1->fetchrow_array) {
		if($self->{sImProtocol}) {
			$sToUser = substr($sToUser, 4);
		}

		$dbh->do("UPDATE IM_EVENT SET new='N' WHERE id=$nEventID");

		if($sType eq "SYS") {
			push(@aEventsToDelete, $nEventID);
		}

		my $sEvent = $self->getCharset()."|$sToUser|$sType|$sCMD|$sDate\n$sAttr";
		push(@{$self->{_aOutQueue}}, $sEvent);
	}
	$sth1->finish();
	# delete fetched SYS events
	if(@aEventsToDelete) {
		$dbh->do("DELETE FROM IM_EVENT WHERE id in (".join(",",@aEventsToDelete).")");
	}
}

sub hasIncomingEvents{
	my $self = shift;
	return 0;
}

sub keepAlive { # we don't need to keep open connection to the IM proxy server
	return 0;
}

sub brokenConnection {
	my $self=shift;

	my $sEventAttr = "    ".$self->{sImProtocol};
	$sEventAttr = "TYPE=ALERT&CONTENT=".$self->_urlencode($sEventAttr)."&CL=0";

	$self->_sendEvent('PRT', 'CONNECTED', $self->{nHostUserID},$self->{sImProtocol}, 0, time());
	$self->_sendEvent('PRT', 'STAT', $self->{nHostUserID},$self->{sImProtocol}, "OFF", time());
	$self->_sendEvent('SYS', 'POP_UP', $self->{nHostUserID}, $self->{sImProtocol}, $sEventAttr, time);
}

sub shutdown {
	my $self = shift();
	my $sEventAttr = ",     ".$self->{sImProtocol}.",   !";
	$sEventAttr = "TYPE=ALERT&CONTENT=".$self->_urlencode($sEventAttr)."&CL=0";
	$self->_sendEvent('SYS', 'POP_UP', $self->{nHostUserID}, $self->{sImProtocol}, $sEventAttr, time);
}

#:: private functions #


sub _sendEvent {
	my $self = shift();
	my ($sType, $sCMD, $nToID, $nFromID, $hAttr, $dDate, $dExpire, $bUpdateInbox) = @_;
	$bUpdateInbox=1 if (!defined($bUpdateInbox));
	$dExpire = 0 if (!defined($dExpire));
	$hAttr ='' if (!defined($hAttr));

	# [don't send messages to myself] #
	if($nFromID eq $nToID) {
		return 0;
	}

	$self->_debug("_sendEvent(): $sType, $sCMD, $nToID, $nFromID, $hAttr, $dDate, $dExpire, $bUpdateInbox");

	if($sType eq 'SYS') {
		$hAttr =~s/'/\\'/g;
		my $sth1 = $dbh->prepare("SELECT id FROM IM_EVENT WHERE type='SYS' AND cmd='$sCMD' AND to_usr= '$nToID' AND from_usr='$nFromID' AND attr='$hAttr' ") or die "Couldn't prepare statement: " . $dbh->errstr;
		$sth1->execute();
		if($sth1->rows()) {
			$sth1->finish();
			return 0;
		}
		$sth1->finish();
	}

	$dbh->do("INSERT INTO IM_EVENT(type, cmd, to_usr, from_usr, attr, date, expire) VALUES ('$sType', '$sCMD',  ?, ?, ?, FROM_UNIXTIME($dDate), '')", undef, $nToID, $nFromID, $hAttr);

	if($bUpdateInbox) {
		$dbh->do("UPDATE IM_SESS SET in_box=in_box+1 WHERE usr=?", undef, $nToID);
	}
	return 1;

}

# [configuration specific functions] #
sub _trim {
    my $self = shift;
    my @out = @_;
    for (@out) {
        s/^\s+//;
        s/\s+$//;
    }
    return wantarray ? @out : $out[0];	
}

sub _loadConfig { #use common config file for PHP, Perl and other scripts
	my ($self, $file) = @_;
	my %hConfig = ();
	if(open(CFG, "< $file")) {
		while(<CFG>) {
			my $line = $self->_trim($_);
			if($line =~ /^#/) {
				next;
			}
			my ($name, $value) = split(/=/,$line);
			if(length($name) && length($value)) {
				$name = $self->_trim($name);
				$value = $self->_trim($value);
				$hConfig{$name}=$value;
			}
		}
		close(CFG);
	} 
	else {
		$self->_debug("Unable to load config \"$file\"");
	}
	return %hConfig;
}

sub END {
	#ToDo:: send information to client that we are exiting if the connection is active
}

1;
