package IM::AIM;
use base 'IM::Base';
#######################################
#	Class for AIM TOC protocol
#	Copyright: Slavej Karadjov -2003
#######################################
use strict;
use Data::Dumper; # for debug purposes

# look at: http://gridley.acns.carleton.edu/~straitm/final/PROTOCOL for protocol description

my $nomodule = 0; 
eval ("require '/home/s_karadjov/workplace/Net/AIM.pm'"); # for debug purposes
if($@) {
	$nomodule++;
	$@ = '';
}
if($nomodule) {
	eval ("use Net::AIM");
	if($@) {
		print "Can't find Net::AIM\n";
		exit;
	}
}
$|=1;

# -- Connection Options -- #
my $nLastMessageTo = 0; # values are UIN
my $oAIM = {};

# [Constructor] #
sub new {
	my $proto = shift;
	my $class = ref($proto) || $proto;
	my $self = $class->SUPER::new(@_);
	$self->{nPingTimeOut} =  110;
	$self->{PROTOCOL} = "AIM";
	$self->{CONTACT_LIST_READY} =0;
	$self->{AUTO_RECONNECT} = 0; # set this to 1 but be warned
	$self->{SOCKET} = undef;
	# ...     AIM  

	$oAIM = new Net::AIM;
	# uncomment the line below if you debug
	# $oAIM->debug(1); 

        bless ($self, $class);
	return $self;
}

# :: public functions
sub connect {
	my $self = shift();
	$self->{CLIENT_IP} = shift;
	$self->_debug("Connecting to AIM  server");   		

	if(!$oAIM->newconn(  Screenname => $self->{USER_ID} ,
	 				Password => $self->{USER_CREDENTIALS},
	 				AutoReconnect => $self->{AUTO_RECONNECT}
      				    	)
	) {
		$self->_debug("connect(): Can't connect to AIM server");			    
		return 0;
	} 	
	if($self->{SOCKET} =  $oAIM->getconn()) {
		$self->{SOCKET}->set_handler('error', sub { return $self->_onError(@_); });
		$self->{SOCKET}->set_handler('im_in', sub { return $self->_receiveMessage(@_); });
		$self->{SOCKET}->set_handler('nick',  sub { return $self->_onNick(@_); });
		$self->{SOCKET}->set_handler('update_buddy', sub { return $self->_updateBuddy(@_); });
		$self->{SOCKET}->set_handler('config', sub { return $self->_onConfig(@_); });
		$self->{SOCKET}->set_handler('eviled', sub { return $self->_onEviled(@_); });
		#$self->{SOCKET}->set_handler('chat_join',    \&on_chat_join);
		#$self->{SOCKET}->set_handler('chat_left',    \&on_chat_left);
		#$self->{SOCKET}->set_handler('chat_in',    \&on_chat_in);
		#$self->{SOCKET}->set_handler('chat_invite',    \&on_chat_invite);
		#$self->{SOCKET}->set_handler('chat_update_buddy',    \&on_chat_update_buddy);
		
		return 1;
	}	
	return 0;	
}
	

sub error {
	my $self=shift;
	my $sText = shift;
	
	#my %ERROR_CODES = (
	#		911=>"Authentication failed", #   -      AIM
	#	);
	# ...
	#$self->_debug("::AIM Error::".$sText."->".$ERROR_CODES{$sText});
	$self->_debug("::AIM Error:: $sText");
	exit();
}

sub isConnected {
	my $self = shift();
	#       AIM /      
	return $self->{SOCKET}->connected();	
}

sub getContactList {
	my $self = shift();

	$self->_debug("getContactList(): Retreiving Contact List");

	# get all users and user groups from remote contact list
	while(!$self->{CONTACT_LIST_READY}) {
		print "+";
		$oAIM->do_one_loop();
	}	
	# get information for the user from contact list -- may be this is it
	
	$self->_debug("getContactList(): Start generating contact list");			
	return $self->_generateContactListString();
}

sub pingIMServer {
	my $self = shift();
	#  PING  
	$self->_debug("pingIMServer(): do ping");
	#$oAIM->ping();
	$self->{nLastPing} = time();
}

sub changeState {
	my($self, $sStatusCode) = @_;

	my %STATUS_IDLE_TIME = (
		'ONLINE' => 0,
		'IDLE' => 30,
		'AWAY' => 120
		);
	#        .       ONLINE, OFFLINE  HIDDEN
	if($sStatusCode eq "ONLINE") {
		$oAIM->set_away($STATUS_IDLE_TIME{$sStatusCode});
	}
	$self->_debug("Status Changed to: $sStatusCode");
}


sub _getProtocolStatus{
	my ($self, $sStatus) = @_;
	my $sNewStatus = undef;
	if($sStatus eq "ONL") {
		$sNewStatus = "ONLINE";
	}
	elsif($sStatus eq "HDN") {
		$sNewStatus = "AWAY";
	}
	elsif($sStatus eq "IDL") {
		$sNewStatus = "AWAY";
	}
	elsif($sStatus eq "BSY") {
		$sNewStatus = "AWAY";
	}
	return $sNewStatus;
}

sub getIncomingEvents {
	my $self = shift;
	#      AIM 
	$oAIM->do_one_loop();
}

sub hasIncomingEvents {
	return 0;
}

sub sendTextMessage {
	my ($self, $toAID, $sMessage) = @_;
	$nLastMessageTo = $toAID;
	$oAIM->send_im($toAID, $sMessage);
	$self->_debug("_sendTextMessage(".${$self->{_hUsers}}{$toAID}->{'NAME'}." [$toAID] ): $sMessage");
	$self->{nLastPing} = time();
}

sub disconnect {
	my($self) = shift();
	$self->_debug("Disconnecting From AIM Server");
	if($self->isConnected()) {
		#   
		$oAIM->disconnect();
	}
	#exit();
}


# ::private functions

# Handle Functions
sub _receiveMessage {
   my ($self, $aimObject, $evt, $from, $to) = @_;
   my $args = $evt->args();
   my ($nick, $auto_msg, $msg) = @$args;

   chomp($msg);
   my $stripped = $msg;
   $stripped =~ s/<[^>]+>//g;
   $stripped =~ s/^\s+//g;

   $self->_debug("_receiveMessage(): Got Message! $nick wrote  $stripped");
#warning:: not sure about the encoding
   $self->_addEventToOutQueue('utf8', $nick, "USR", "NEW_MSG", $stripped);
}

sub _onNick {
	my ($self, $aimObject, $evt, $from, $to) = @_;
	my $args = $evt->args();
	my $nick = $args->[0];
#NICK:<Nickname>
#   Tells you your correct nickname (ie how it should be capitalized and  spacing)
	if(exists(${$self->{_hUsers}}{$nick})) {
		# hey - thats me - don't add me to buddy list
		#delete($hUser{$nick});		
	}
	$self->_debug("_onNick(): $nick");
}

sub _updateBuddy {
   my ($self, $aimObject, $evt, $from, $to) = @_;
   my ($nick) = $from;
   my ($bud, $online, $evil, $signon_time, $idle_amount, $user_class) = @{$evt->args()};

   #lets print out when users leave and when they enter;
   if ($online eq 'T') {
	#$hUser{$bud}=$bud;
	#ToDo:: add new online users ....
#warning:: something wrong with this -> fix it
	if(!exists(${$self->{_hUsers}}{$bud}))  {
			${$self->{_hUsers}}{$bud}->{NAME} = $bud;
#warning:: not sure about the encoding
			$self->_addEventToOutQueue('utf8', $bud, "SYS", "ADD_USER", $bud);
	}
	$self->_debug("_updateBuddy(): $bud has signed on at " . scalar localtime());
	$self->_addEventToOutQueue('cp1251', $bud, "SYS", "STAT", 'ONL');
   }
   elsif ($online eq 'F') {
   	$self->_debug("_updateBuddy():  $bud has signed off at " . scalar localtime());
      	$self->_addEventToOutQueue('cp1251', $bud, "SYS", "STAT", 'OFF');
   }
}

sub _onConfig {
   my ($self, $aimObject, $evt, $from, $to) = @_;

   my $str = shift @{$evt->args()};
   $aimObject->set_config_str($str, 1);

   foreach my $group (keys %{ $aimObject->{_config}} ) {
   	next if($group eq "mode");
   	${$self->{_hGroups}}{$group} = $group;
	while (my ($sn, $type) = each %{ $aimObject->{_config}->{$group} } ) {
		if($type eq "b") { # we have buddy
			${$self->{_hUsers}}{$sn}->{'NAME'} = $sn;
			if(${$self->{_hUsers}}{$sn}->{'GROUP_ID'}) {
				${$self->{_hUsers}}{$sn}->{'GROUP_ID'} .= ",$group";
			}
			else {
				${$self->{_hUsers}}{$sn}->{'GROUP_ID'} = "$group";
			}
		}
   	}
   }
   $self->{CONTACT_LIST_READY} = 1;

   $aimObject->send_buddies();
   $aimObject->send_config();
}

sub _onError {
   my ($self, $aimObject, $evt) = @_;
   my ($error, @stuff) = @{$evt->args()};

   my $errstr = $evt->trans($error);
   $errstr =~ s/\$(\d+)/$stuff[$1]/ge;

   # error handling
   if($error == 983) { # connecting too fast
   	$aimObject->disconnect();
   }
   elsif($error == 901) { # XXX not currently available
   	$self->_debug("_onError(): user ".$stuff[0]." is offline");
	$self->_addEventToOutQueue('cp1251', $stuff[0], "SYS", "STAT", 'OFF');
   }
   $self->_debug("_onError(): $errstr");
}

sub _onEviled { 
	my ($self, $aimObject, $evt, $from, $to) = @_;
   	my ($level, $culprit) = @{$evt->args};
	$culprit = 'An anonymous user' if ($culprit =~ /^\s*$/);
   	$self->_debug("_onEviled(): $culprit slapped us! Our evil level is now $level");
}


=head Sample Handlers
sub on_chat_left {
   my ($self, $evt) = @_;

   my ($id) = @{$evt->args()};
   my $name = $self->get_roomname($id);

   $self->del_roomname($id);
   print "We left room $name\n";

}


sub on_chat_join {
   my ($self, $evt) = @_;
   my ($id, $name) = @{$evt->args()};

   $self->set_roomname($id, $name);
   print "We joined room $name\n";
}


sub on_chat_update_buddy {
   my ($self, $evt) = @_;
   my ($id, $inside, @users) = @{$evt->args()};

   my $room = $self->get_roomname($id);
   my %act = (
   	'T' => 'entered',
   	'F' => 'left'
   );

   foreach my $u (@users) {
      print "$u has $act{$inside} $room\n";
   }
}


sub on_chat_in {
   my ($self, $evt) = @_;
   my ($id, $user, $whisper, $msg) = @{$evt->args()};

   my $action = 'said';
   $action = 'whispered' if ($whisper =~ /T/i);

   print $self->get_roomname($id) . "> $user $action $msg\n";
	
   if ($msg =~ /please leave/i) {
      $self->chat_leave($id);
   }
}

sub on_chat_invite {
   my ($self, $evt) = @_;

   my ($name, $id, $user, $msg) = @{$evt->args()};
   print "Invited to room $name by $user\n";
   $self->chat_accept($id);
}

sub on_eviled {
   my ($self, $evt, $from, $to) = @_;
	
   my ($level, $culprit) = @{$evt->args};
#	$culprit = 'An anonymous user' if ($culprit =~ /^\s*$/);
   print "$culprit slapped us! Our evil level is now $level\n";

   #should we hit them back twice??
   
   if ($culprit !~ /^\s*$/) {
      $morons{$self->normalize($culprit)}++;

      my $nick = $self->normalize($culprit);
      if ( exists $morons{$nick}) {
	 $self->evil($nick);
	 sleep(1);
	 $self->evil($nick);
	 sleep(1);
	 $self->evil($nick);
#	 $self->send_im('myscreenname', 'I warned ' . $nick);
	 return;
      }

#      $self->send_im('myscreenname', $culprit . ' eviled me that moron!');
      $self->evil($culprit);
      $self->evil($culprit);
      $self->send_im($culprit, 'Yo what\'s your problem?');

   } else {
     my $lmd =  $self->{_last_msg_dest};
     if (++$suspects{$lmd} > 3) {
        $morons{$lmd}++;
        $self->evil($lmd);
     }

   }
}
=cut




1;

__END__
