package IM::Base;
##############################################
#	Base class for IM protocols
#	Copyright: Slavej Karadjov -2003
##############################################
use strict;

sub new {
	my $proto = shift;
	my $class = ref($proto) || $proto;
	my $self = {
			USER_ID => shift,
			USER_CREDENTIALS => shift,
			SCREEN_NAME =>undef,
			CLIENT_IP => undef,
			PROTOCOL => undef,
			STATUS => 'OFF',
			_CONNECTED => 0,
			nLastPing => 0,
			nPingTimeOut => 50, #               
			_aOutQueue => [],
			_aInQueue => [],
			_hGroups => {},
			_hUsers => {},
		    };
	bless ($self, $class);
}

sub connect {
	my $self = shift;
	$self->{CLIENT_IP} = shift;
	$self->_debug("connect(): NOT IMPLEMENTED");
	$self->{_CONNECTED} = 1;
}

sub changeState {
	my $self = shift;
	$self->_debug("changeState(): NOT IMPLEMENTED");
}

sub isTimeToPingIMServer {
	my $self = shift();
	if(time() - $self->{nLastPing} >  $self->{nPingTimeOut}) {
		return 1;
	}
	return 0;
}


sub pingIMServer {
	my $self = shift;
	$self->_debug("pingIMServer(): NOT IMPLEMENTED");
}

sub isConnected {
	my $self = shift;
	return $self->{_CONNECTED};
}

sub getContactList {
	my $self = shift;
	$self->_debug("getContactList(): NOT IMPLEMENTED");
}
# [User and ContactList functionality] #
sub addUser {
	my ($self,$sUserID) = @_;
	$self->_debug("addUser(): NOT IMPLEMENTED");
}

sub delUser {
	my ($self,$sUserID) = @_;
	$self->_debug("delUser(): NOT IMPLEMENTED");
}

sub renameUser {
	my ($self,$sUserID, $sUserName) = @_;
	$self->_debug("renameUser(): NOT IMPLEMENTED");
}
# [Group and ContactList functionality] #
sub addGroup {
	my ($self,$sGroupID, $sGroupName) = @_;
	$self->_debug("addGroup(): NOT IMPLEMENTED");
}

sub delGroup {
	my ($self,$sGroupID) = @_;
	$self->_debug("delGroup(): NOT IMPLEMENTED");
}

sub renameGroup {
	my ($self,$sGroupID, $sGroupName) = @_;
	$self->_debug("renameGroup(): NOT IMPLEMENTED");
}

sub moveUserToGroup { # Add/Move User To Group
	my ($self,$sUserID, $sFromGroupID, $sToGroupID) = @_;
	$self->_debug("moveUserToGroup(): NOT IMPLEMENTED");
}

sub delUserFromGroup {
	my ($self,$sUserID, $sGroupID) = @_;
	$self->_debug("moveUserToGroup(): NOT IMPLEMENTED");
}

# [Authorization/Permission section] #
sub grantAuth {
	my ($self,$sUserID,$bAllow) = @_;
	$self->_debug("grantAuth(): NOT IMPLEMENTED");
}

sub sleep {
	#         ,
	#         ,     
	CORE::sleep(2);
}

sub getIncomingEvents {
	my $self = shift;
	$self->_debug("getIncomingEvents(): NOT IMPLEMENTED");
}

sub getIncomingEventsFromQueue {
	my $self = shift;
	my @aResult = @{$self->{_aInQueue}};
	@{$self->{_aInQueue}} = ();
	return @aResult;
}

sub hasIncomingEvents {
	#      IM 
	return 0;
}

sub hasOutgoingEvents {
	my $self = shift;
	$self->getIncomingEvents();
	return @{$self->{_aInQueue}};
}

sub sendTextMessage {
	my ($self, $sTo, $sText) = @_;
	#           IM   IM 
	$self->_debug("sendTextMessage(): NOT IMPLEMENTED");
}

sub sendOutgoingEvents {
	my $self = shift();
	@{$self->{_aInQueue}} = @_;
	foreach(@{$self->{_aInQueue}}) {
		my $sEventHeader = substr($_,0, index($_, "\n")-1);
		my $sEventAttr = substr($_, index($_, "\n")+1);

		my ($enc, $to, $type, $cmd, $attr) = split(/\|/, $sEventHeader);

		if($type eq "USR" && $cmd eq "NEW_MSG") {
			#ToDo: convert message from $enc to utf8;
			$self->sendTextMessage($to, $sEventAttr);
		}
		elsif($type eq "SYS") {
			if($to eq "SERVER") {
				if($cmd eq "STAT") {
					# Translate Statuses to local system
					my $sNewPresenceStatus = $self->_getProtocolStatus($sEventAttr);

					if($sNewPresenceStatus) {
						$self->changeState($sNewPresenceStatus);
					}
				}
				elsif($cmd eq "DISCONNECT") {
					$self->disconnect();
				}
				elsif($cmd eq "ADDU") {
					$self->addUser($sEventAttr);
				}
				elsif($cmd eq "DELU") {
					$self->delUser($sEventAttr);
				}
				elsif($cmd eq "RENU") {
    					my %hData = $self->_parseAttrData($sEventAttr);
					$self->renameUser($hData{'USER_ID'},$hData{'USER_NAME'});
				}
				#ToDo: .... add the other IM server calls here
			}
			else { # if the message is not directed to the server
				if($cmd eq "GRANT_AUTH") {
					$self->grantAuth($to,1);
					$self->addUser($to);
				}
			}
		}
	}
	@{$self->{_aInQueue}} = ();
}

sub disconnect {
	my $self = shift();
	$self->_debug("disconnect(): NOT IMPLEMENTED");
	$self->{_CONNECTED} = 0;
}

sub error {
	my ($self,$sText) = @_;
	$self->_debug("error():: $sText");
	exit(1);
}


# private functions
sub _debug {
	my($self, $sText) = @_;
	print "IM::".$self->{PROTOCOL}." module [".$self->{CLIENT_IP}." - ".(scalar localtime())."] : $sText\n";
}

sub _error {
	my($self, $sText) = @_;
	print "IM::".$self->{PROTOCOL}." module error [".$self->{CLIENT_IP}." - ".(scalar localtime())."] : $sText\n";
	exit(1);
}

sub _addEventToOutQueue {
	my ($self, $enc, $from, $type, $cmd, $attr) = @_;
	# ToDo:: escape delimiter "|" in  attr and from
	my $sEvent = "$enc|$from|$type|$cmd|".time()."\n$attr";
	$self->_debug("_addEventToOutQueue(): [$sEvent]");
	push(@{$self->{_aInQueue}}, $sEvent);
}

sub _generateContactListString {
	my $self = shift();
	my $sContactList = "utf8;[";
	my $sGroups = "";
	foreach(keys(%{$self->{_hGroups}})) {
		# generate list of groups
		$sGroups.=${$self->{_hGroups}}{$_}.",".$_."|";
	}
	if($sGroups) {
		$sContactList .=substr($sGroups,0, -1);
	}
	$sContactList.="];[";
	my $sUsers = "";
	
	foreach(keys(%{$self->{_hUsers}})) {
		# generate list of users
		$sUsers.=$_.",".${$self->{_hUsers}}{$_}->{'NAME'}.",{".${$self->{_hUsers}}{$_}->{'GROUP_ID'}."}|";
	}
	if($sUsers) {
		$sContactList .=substr($sUsers,0, -1);
	}
	$sContactList.="]";
	# clear users and groups hashes
	%{$self->{_hGroups}} = {};
	%{$self->{_hUsers}} = {};
	return $sContactList;
}

sub _getProtocolStatus {
	my $self = shift;
	$self->_debug("_getProtocolStatus(): NOT IMPLEMENTED");
}

sub _urldecode {
	my ($self, $sText) = @_;
	$sText =~ tr/+/ /;
        $sText =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
	return $sText;
}

sub _parseAttrData {
	my ($self, $sText) = @_;
	my %hData = ();
	my @tPairs = split(/\&/,$sText);
	foreach(@tPairs) {
		my($tName, $tValue) =split(/=/);
		$hData{uc($tName)} = $self->_urldecode($tValue);
	}
	return %hData;
}

1;
__END__
=head  /         IM Proxy-
   :

ENCODING|USER_ID|TYPE|CMD|TIME\n
ATTRIBUTES

   HEADER  ,       ATTRIBUTES

ENCODING -         ATTRIBUTES 
USER_ID -   AIM ID-    
TYPE -    . -  SYS - /  USR -   
CMD -       .       
	  SYS      
		LOG_IN
		LOG_OUT
	 USR    
		NEW_MSG
ATTRIBUTES -     .

             AIM.
:

utf8|43434343|SYS|LOG_IN|10000000\n
Mironcho

   
utf8|43434343|SYS|LOG_OUT|10000000\n

  		
utf8|43434343|USR|NEW_MSG|10000000\n
Towa e nowo syobshtenie kodirano w utf8


=cut
