package IM::Base;
##############################################
#	Class for Gadu IM protocol
#	Copyright: Slavej Karadjov -2003
##############################################
use strict;
use Net::Gadu;

my $oGG = {}; #GaduGadu object
sub new {
	my $proto = shift;
	my $class = ref($proto) || $proto;
	my $self = $class->SUPER::new(@_);
	$self->{nPingTimeOut} =  50;

	$self->{PROTOCOL} = "GADU-GADU";
	$self->{IS_CONNECTED}=0;
        $self->{CONTACT_LIST_READY} =0;

	$oGG = new Net::Gadu(async=>1);

	bless ($self, $class);
	return $self;
}

sub connect {
	my $self = shift;
	$self->{CLIENT_IP} = shift;
	$self->_debug("connect(): Connecting to GaduGadu server ...");
	while(1) {
		print "*";
		$oGG->login($self->{'USER_ID'},$self->{'USER_CREDENTIALS'}) && last;

		$self->_debug("connect(): Invalid Login. Make sure your ID and PASSWORD are entered correctly.");
		return 0;
	}
	return 1;
}

sub changeState {
	my ($self,$sStatus) = shift;
	if($sStatus eq "ONLINE") {
		$oGG->set_available();
	}
	elsif($sStatus eq "BUSY") {
		$oGG->set_busy();
	}
	elsif($sStatus eq "IDLE") {
		$oGG->set_not_available();
	}
	elsif($sStatus eq "HIDDEN") {
           	$oGG->set_invisible();
	}
	$self->_debug("changeState(): to $sStatus");
}

sub pingIMServer {
	my $self = shift;
	$oGG->ping();
}

sub isConnected {
	my $self = shift;
	return $self->{IS_CONNECTED};
}

sub getIncomingEvents {
	my $self = shift;
	if ($oGG->check_event() == 1) {
        	my $e = $oGG->get_event();

        	my $type = $e->{type};

		$self->_debug("getIncomingEvents(): Got Event:".$type);

        	if ($type == $Net::Gadu::EVENT_CONN_SUCCESS) {
			$self->{IS_CONNECTED} = 1;
        	}

        	if ($type == $Net::Gadu::EVENT_MSG) {
			$self->_addEventToOutQueue('cp1251', $e->{sender}, 'USR', 'NEW_MSG', $e->{message});
        	}

        	if ($type == $Net::Gadu::EVENT_ACK) {
			# ACK received
        	}

		if($type == $Net::Gadu::EVENT_STATUS) {
			$self->_addEventToOutQueue('cp1251', $e->{sender}, 'SYS', 'STAT', $e->{status});
		}

		if($type == $Net::Gadu::EVENT_NOTIFY) {
			# not sure about this one
			$self->_debug("getIncomingEvents(): EVENT_NOTIFY");
		}
		if($type == $Net::Gadu::EVENT_NOTIFY_DESCR) {
			# not sure about this one
			$self->_debug("getIncomingEvents(): EVENT_NOTIFY_DESCR");
		}

		if($type == $Net::Gadu::EVENT_CONN_FAILED) {
			$self->_debug("getIncomingEvents(): Connection Failed");
			$self->_addEventToOutQueue('cp1251', -1, 'SYS', 'DISCONNECT', "Connection Failed");
		}

		if($type == $Net::Gadu::EVENT_DISCONNECT) {
			# received disconnect from server
			$self->_addEventToOutQueue('cp1251', -1, 'SYS', 'DISCONNECT', "Received disconnect from server");
		}
      }
}


sub sendTextMessage {
	my ($self, $sTo, $sText) = @_;
	#           IM   IM 
	$oGG->send_message($sTo, $sText);
}


sub disconnect {
	my $self = shift();
	$self->_debug("disconnect(): Disconnecting from Gadu-Gadu Servers");
	$oGG->logoff();
}

sub error {
	my ($self,$sText) = @_;
	$self->_debug("error():: $sText");
	exit(1);
}

# private functions
sub _getProtocolStatus {
	my $self = shift;
	$self->_debug("_getProtocolStatus(): NOT IMPLEMENTED");
}

1;
