package IM::MSN;
#BEGIN {
        # we need this for the Net::SSLeay
#       $ENV{'LD_LIBRARY_PATH'}.=":/home/linuxbg/openssl/lib/";
#}
use base 'IM::Base';
#######################################
#       Class for new MSN protocols
#       Copyright: Slavej Karadjov -2003
#######################################
use strict;
use Socket;
use IO::Socket;
use IO::Handle;
use Digest::MD5 qw/md5_hex/;
use Net::SSLeay qw(get_https post_https sslcat make_headers make_form);
#warning:: for debug purposes only
#use Data::Dumper;

BEGIN {
        eval("use Encode qw(from_to);");
        if(@!) {
                sub from_to($$$;$) {
                        return @_;
                }
        }
}

# Use high-res timing if avaliable
#BEGIN { eval 'use Time::HiRes qw/time/' } # -- we don't need it for now

# [configuration ] #

# -- MSN settings -- #
my $sDSServerAddress = 'messenger.hotmail.com'; # if value is undef - use default settings
my $nDSServerPort = 1863; # if value is undef - use dedault settings

my $sNSServerAddress = undef;
my $nNSServerPort = undef;

my $sSSServerAddress = undef; # switchboard server
my $nSSServerPort = undef;

# -- Chat User Sessions -- #
my %hChatSess = ();

# [Constructor] #
sub new {
        my $proto = shift;
        my $class = ref($proto) || $proto;
        my $self = $class->SUPER::new(@_);
        $self->{nPingTimeOut} =  50;

        $self->{_nUsersToFetch} = -1;
        $self->{_nGroupsToFetch} = -1;

        $self->{PROTOCOL} = "MSN8";
        $self->{SOCKET}  = undef;
        $self->{SB_SOCKET}  = undef; # SwitchBoard Socket
        $self->{TRANS_ID} = 0; # transaction ID
        $self->{SB_TRANS_ID} = 1; # SwitchBoard TransactionID
        $self->{UNREAD_EMAILS}  = undef;
        $self->{AUTH_STRING} = undef;
        $self->{SOCKET_IS_NONBLOCKING} = 0;

        bless ($self, $class);
}

# :: public functions
sub connect {
        my $self = shift();
        $self->{CLIENT_IP} = shift;
        $self->{USER_ID} = $self->_fixUserID($self->{USER_ID});
        $self->_debug("Connecting to MSN Dispatch server");
        ($sNSServerAddress, $nNSServerPort) = $self->_connectToDispatchServer();

        if(!$sNSServerAddress || !$nNSServerPort) {
                $self->_debug("Unable to connect do Notification server. Exiting");
                return 0;
        }

        if($self->_connectToAuthenticationServer($sNSServerAddress, $nNSServerPort)) {
                $self->_debug("Ou Key - We are connected to notification server.");
                $self->{'_CONNECTED'} = 1;
        }
        else {
                return 0;
        }

        return 1;
}


sub error {
        my $self=shift;
        my $sText = shift;

        my %ERROR_CODES = (
                        911=>"Authentication failed",
                        540 =>"Challenge response failed"
                );

        $self->_debug("::MSN Error::".$sText."->".$ERROR_CODES{$sText});
        #exit();
}

sub getContactList {
        my $self = shift();
        my $buf = undef;

        $self->_debug("Retreiving Contact List");
        $self->_send($self->_genTransCmd("SYN",0));
        # The second parameter is the number of LST responses to expect.
        # The third parameter is the number of LSG responses to expect.
        #  <<< SYN 1 139 6 4
        while(my $buf = $self->_recv()) {
                my @bit = split(/ /,$buf);
                if($bit[0] eq "SYN") {
                        $self->{_nUsersToFetch} = $bit[3];
                        $self->{_nGroupsToFetch} = $bit[4];
                        last;
                }
        }

        $self->_makeSocketNonBlocking();
        #while(!($self->{_nUsersToFetch}==0 && $self->{_nGroupsToFetch}==0)) {
        while($self->{_nUsersToFetch}>0) {
                $buf=$self->_recv();
                if(!$buf) {
                        next;
                }
                $self->_debug('Users: '.$self->{_nUsersToFetch}.', Groups:'.$self->{_nGroupsToFetch});
        }
        $self->_debug("getContactList(): Start fetching contact list");
        return $self->_generateContactListString();
}

sub pingIMServer {
        my $self = shift();
        $self->_send("PNG");
        my $buf = $self->_recv();
        $self->{nLastPing} = time();
}

sub changeState {
        my($self, $sStatusCode) = @_;

        my %STATUS = (
                        "ONLINE"=>"NLN",
                        "OFFLINE"=>"FLN",
                        "HIDDEN"=>"HDN",
                        "BUSY"=> "BSY",
                        "IDLE"=> "IDL"
                );

        if($STATUS{$sStatusCode} eq "") {
                #$self->_error("Invalid Status code");
                $self->_debug("Invalid Status code");
        }

        $self->_send($self->_genTransCmd("CHG",$STATUS{"$sStatusCode"}));
#ToDo:: may be we don't need to take care of this response
        my $buf=$self->_recv();
        if("CHG" eq (split(/ /,$buf))[0]) {
                $self->_debug("Status Changed to: $sStatusCode");
        }
        else {
                $self->_debug("Unable to change status. Error was: $buf");
        }
}

sub getIncomingEvents {
        my $self = shift;
        $self->_recv();
        $self->_recvAllFromSB();
}

sub _getProtocolStatus{
        my ($self, $sStatus) = @_;
        my $sNewStatus = undef;
        if($sStatus eq "ONL") {
                $sNewStatus = "ONLINE";
        }
        elsif($sStatus eq "HDN") {
                $sNewStatus = "HIDDEN";
        }
        elsif($sStatus eq "IDL") {
                $sNewStatus = "IDLE";
        }
        elsif($sStatus eq "BSY") {
                $sNewStatus = "BUSY";
        }
        return $sNewStatus;
}

sub sendTextMessage {
        my($self, $to, $text) = @_;
#warning:: experimental stuff
        from_to($text,"cp1251","utf8");
        $self->_sendSBMessage($to, $text);
}

sub disconnect {
        my($self) = shift();
        $self->_debug("Disconnecting SB sessions");
        foreach my $sUserID (keys(%hChatSess)) {
                if(!fileno($hChatSess{$sUserID}->{'SB_SOCKET'}) || !defined($hChatSess{$sUserID}->{'SB_SOCKET'})) {
                        # don't send BYE messages to closed socket
                        next;
                }
                $self->_sendToSB($sUserID, "BYE $sUserID");
                close($hChatSess{$sUserID}->{'SB_SOCKET'});
                $hChatSess{$sUserID}->{'SB_SOCKET'} = undef;
        }

        $self->_debug("Disconnecting From MSN Server");
        if($self->isConnected()) {
                $self->_send('OUT');
                close($self->{SOCKET});
                $self->{SOCKET} = undef;
                $self->{'_CONNECTED'} = 0;
        }
        #exit();
}

# ::private functions

# [error/status/debug functions] #
sub _fixUserID {
        my ($self, $user) = @_;

        $user =~ /^([^@]+)(.*)/;

        lc $1 . ($2 || '@hotmail.com')
}

#  [server connection section] #
sub _connectToDispatchServer {
        my $self = shift();

        $self->{TRANS_ID}=0;

        # -- create socket
        $self->{SOCKET} = new IO::Handle;
        socket($self->{SOCKET}, PF_INET, SOCK_STREAM, getprotobyname 'tcp') or  die("socket() returned $!");

        # -- connect socket to Dispatch Server
        if(!CORE::connect $self->{SOCKET}, sockaddr_in($nDSServerPort, inet_aton($sDSServerAddress))
                                        and $! ne 'Operation now in progress') {
                undef $self->{SOCKET};
                $self->_error("connect() returned $!");
        }
        $self->{SOCKET}->autoflush(1);

        $self->_send($self->_genTransCmd("VER","MSNP8 CVR0"));
        $self->_recv();

        # CVR 1 0x0409 winnt 5.1 i386 MSNMSGR 5.0.0 MSMSGS
        $self->_send($self->_genTransCmd("CVR","0x0409 win 4.10 i386 MSNMSGR 5.0.0544 MSMSGS ".$self->{USER_ID}));

        $self->_recv();

        $self->_send($self->_genTransCmd("USR","TWN I ".$self->{USER_ID}));
        my $sResult = $self->_recv();

        # close socket
        close($self->{SOCKET});
        undef $self->{SOCKET};

        # XFR 3 NS 207.46.106.35:1863 0 207.46.104.20:1863
        my @bit= split(/ /,$sResult);
        return split(/:/,$bit[3]); # Server And Port
}

sub _connectToAuthenticationServer {
        my ($self, $server, $port) = @_;

        $self->{TRANS_ID}=1;

        # -- create socket
        $self->{SOCKET} = new IO::Handle;
        socket($self->{SOCKET}, PF_INET, SOCK_STREAM, getprotobyname 'tcp') or  die("socket() returned $!");

        # -- connect socket to Dispatch Server
        if(!CORE::connect $self->{SOCKET}, sockaddr_in($port, inet_aton($server))
                                        and $! ne 'Operation now in progress') {
                undef $self->{SOCKET};
                $self->_error("connect() returned $!");
        }
        $self->{SOCKET}->autoflush(1);

        $self->_send($self->_genTransCmd("VER","MSNP8 CVR0"));
        $self->_recv();

        $self->_send($self->_genTransCmd("CVR","0x0409 win 4.10 i386 MSNMSGR 5.0.0544 MSMSGS ".$self->{USER_ID}));
        $self->_recv();

        $self->_send($self->_genTransCmd("USR","TWN I ".$self->{USER_ID}));
        my $sResult = $self->_recv();
        my @bit = split(/ /,$sResult);
        # USR 3 TWN S lc=1033,id=507,tw=40,fs=1,ru=http%3A%2F%2Fmessenger%2Emsn%2Ecom,ct=1062764229,kpp=1,kv=5,ver=2.1.0173.1,tpf=43f8a4c8ed940c04e3740be46c4d1619\r\n
        my $sAuthResp = $self->_auhenticateWithPassportServer('nexus.passport.com', 443, $bit[4]);
        if(!$sAuthResp) {
                $self->_debug("No AuthResp");
                return 0;
        }
        $self->_send($self->_genTransCmd("USR","TWN S ".$sAuthResp));
        $self->_recv();

        # close socket
        #close($self->{SOCKET});
        #undef $self->{SOCKET};
        return 1;
}


sub _connectToLoginServer {
        my ($self, $server, $port, $path, $sChallengeKey) = @_;

        my $fixed_user_name = $self->{USER_ID};
        $fixed_user_name =~s/\@/\%40/;
#warning:: see it the password need fixing...

=head
>>> GET /login2.srf HTTP/1.1\r\n
>>> Authorization: Passport1.4 OrgVerb=GET,OrgURL=http%3A%2F%2Fmessenger%2Emsn%2Ecom,sign-in=example%40passport.com,pwd=password,           lc=1033,id=507,tw=40,fs=1,ru=http%3A%2F%2Fmessenger%2Emsn%2Ecom,ct=1062764229,kpp=1,kv=5,ver=2.1.0173.1,tpf=43f8a4c8ed940c04e3740be46c4d1619\r\n
>>> Host: login.passport.com\r\n
=cut
        my ($page, $response, %reply_headers) =
                        get_https($server, $port, $path,
                                make_headers(
                                        Host    => $server,
                                        Authorization=> "Passport1.4 OrgVerb=GET,OrgURL=http%3A%2F%2Fmessenger%2Emsn%2Ecom,sign-in=$fixed_user_name,pwd=".$self->{USER_CREDENTIALS}.",".$sChallengeKey
                                )
                        );

#warning: debug foreach
        foreach(keys(%reply_headers)) {
                print "\t".$_."=".$reply_headers{$_}."\n";
        }

        if($reply_headers{'AUTHENTICATION-INFO'} =~ /^Passport1.4 da-status=redir/) {
                $self->_debug('_auhenticateWithPassportServer(): redirect');

                if($reply_headers{'LOCATION'} =~ m/^https:\/\/(.*?)\/(.*)$/) {
                        $server = $1;
                        $path = "/".$2;
                }
                else {
                        $server = 'loginnet.passport.com';
                        $path = '/login2.srf?lc=1033';
                }
                return $self->_connectToLoginServer($server, $port, $path, $sChallengeKey);
        }
        elsif($reply_headers{'AUTHENTICATION-INFO'}=~ /^Passport1.4 da-status=success/) {
                    $reply_headers{'AUTHENTICATION-INFO'} =~ m/^.+from-PP='(.+)'.+$/;
                    if (my $magic_string = $1) {
                        return $magic_string;
                    }
        }
        else {
                    # WTF
                    $self->_debug('_auhenticateWithPassportServer(): Strange Authentication-Info. This has never happened before.');

        }
        return undef;
}
sub _auhenticateWithPassportServer {
        my ($self, $server, $port, $sChallengeKey) = @_;
        my $sLoginServerURL = "";
        my ($page, $response, %reply_headers) = get_https($server, $port, '/rdr/pprdr.asp');

=head Answer
...
<<< PassportURLs: DARealm=Passport.Net,DALogin=login.passport.com/login2.srf,DAReg=http://register.passport.net/uixpwiz.srf,Properties=https://register.passport.net/editprof.srf,Privacy=http://www.passport.com/consumer/privacypolicy.asp,GeneralRedir=http://nexusrdr.passport.com/redir.asp,Help=http://memberservices.passport.net/memberservice.srf,ConfigVersion=11\r\n
..
=cut
        my @pair = split(/,/,$reply_headers{'PASSPORTURLS'});
        foreach(@pair) {
                 my($key, $val) = split(/=/);
                 if($key eq "DALogin") {
                        $sLoginServerURL = $val
                 }
        }

        my ($loginServer, $loginScript) = ();
        if($sLoginServerURL) {
                $loginServer = substr($sLoginServerURL,0,index($sLoginServerURL,"/"));
                $loginScript = substr($sLoginServerURL,index($sLoginServerURL,"/"),length($sLoginServerURL));
        }

        return $self->_connectToLoginServer($loginServer, $port, $loginScript, $sChallengeKey);
}

sub _connectToSwitchboardServer {
        my $self = shift();
        my @bit = @_;

        # RNG 11752013 207.46.108.38:1863 CKI 849102291.520491113 example@passport.com Example%20Name

        ($sSSServerAddress, $nSSServerPort) = split(/:/,$bit[2]);
        my $sUserID = $bit[5];
        $hChatSess{$sUserID}->{'USER_NAME'}=$bit[6];
        $hChatSess{$sUserID}->{'SB_IP'} = $sSSServerAddress;
        $hChatSess{$sUserID}->{'SB_PORT'} = $nSSServerPort;


        # -- create socket
        my $newSocket = new IO::Handle;
        socket($newSocket, PF_INET, SOCK_STREAM, getprotobyname 'tcp') or  die("socket() returned $!");

        # -- connect socket to Switchboard Server
        if(!CORE::connect $newSocket, sockaddr_in($nSSServerPort, inet_aton($sSSServerAddress))
                                        and $! ne 'Operation now in progress') {
                undef $newSocket;
                $self->_error("connect() returned $!");
        }
        $newSocket->autoflush(1);
        $self->_debug("Connected To Switchobard Server");

#warning:: Make this thing working
        # duplicate socket - not sure it this works
        $hChatSess{$sUserID}->{'SB_SOCKET'} = $newSocket;

        #ANS 1 name_123@hotmail.com 849102291.520491113 11752013
        $self->_sendToSB($sUserID, $self->_genSBTransCmd("ANS",$self->{'USER_ID'}." ".$bit[4]." ".$bit[1]) );
        $self->_recvFromSB($sUserID);

        $hChatSess{$sUserID}->{'SB_READY'}=1;

        #$self->_makeSocketNonBlocking($hChatSess{$sUserID}->{'SB_SOCKET'});
        $hChatSess{$sUserID}->{'SB_SOCKET'}->blocking(0);

        # delete $newSocket socket after it is duplicated ;
}

sub _initiateSwitchboardServerConnection {
        my $self = shift();
        my $sUserID = shift();
        my $sCommand = shift();

        # XFR 10 SB 207.46.108.37:1863 CKI 17262740.1050826919.32308
        my @bit = split(/ /,$sCommand);

        my ($sSSServerAddress, $nSSServerPort) = split(/:/,$bit[3]);
        $hChatSess{$sUserID}->{'USER_NAME'} = $sUserID;
        $hChatSess{$sUserID}->{'SB_IP'} = $sSSServerAddress;
        $hChatSess{$sUserID}->{'SB_PORT'} = $nSSServerPort;

        # -- create socket
        my $newSocket = new IO::Handle;
        socket($newSocket, PF_INET, SOCK_STREAM, getprotobyname 'tcp') or  die("socket() returned $!");

        # -- connect socket to Switchboard Server
        if(!CORE::connect $newSocket, sockaddr_in($nSSServerPort, inet_aton($sSSServerAddress))
                                        and $! ne 'Operation now in progress') {
                undef $newSocket;
                $self->_error("connect() returned $!");
        }
        $newSocket->autoflush(1);
        $self->_debug("Connected To Switchobard Server");
        $hChatSess{$sUserID}->{'SB_SOCKET'} = $newSocket;

        #USR 0 slaff2@hotmail.com 17066187.1054570888.9218
        $self->_sendToSB($sUserID, $self->_genSBTransCmd("USR",$self->{'USER_ID'}." ".$bit[5]) );
        $self->_recvFromSB($sUserID);

        #CAL 2 name_123@hotmail.com
        $self->_sendToSB($sUserID, $self->_genSBTransCmd("CAL",$sUserID));
        $self->_recvFromSB($sUserID);

        #$self->_makeSocketNonBlocking($newSocket);
        $hChatSess{$sUserID}->{'SB_SOCKET'}->blocking(0);
}

sub _sendSBMessage {
        my ($self, $sUserID, $sText)  = @_;

        if(!$hChatSess{$sUserID}->{'SB_READY'}) {
                        $self->_debug("_sendSBMessage($sUserID): still not ready to send command");
#warning:: not sure if this is correct
                        #  add message to Queue
                        push(@{ $hChatSess{$sUserID}->{'InQueue'} }, $sText);

                        if(!defined($hChatSess{$sUserID}->{'SB_SOCKET'}) ) {
                                #return undef;
                                # -- connect to switchboard server and send the message
                                $self->_requestSBSession($sUserID);
                        }
                        return 0; # - don't return - just send message even if we are not ready
        }

        $sText = "MIME-Version: 1.0\r\n".
                        "Content-Type: text/plain; charset=UTF-8\r\n".
                        "X-MMS-IM-Format: FN=Arial; EF=I; CO=0; CS=0; PF=22\r\n".
                        "\r\n".$sText;

        $self->_sendToSB($sUserID, $self->_genSBTransCmd("MSG", "A ".length($sText)."\r\n".$sText),1);
}

# [manipulation functions] #

sub _genTransCmd { # generate transaction command
        my ($self, $sCommand, $sParameters, $nTransactionID, $sTransIDType) = @_;
        $sTransIDType='TRANS_ID'  if(!$sTransIDType);
        my $bIncrementTransactionID = 0;
        if(!$nTransactionID) {
                $nTransactionID = $self->{"$sTransIDType"};
                $bIncrementTransactionID = 1;
        }

        my $sCMD = "$sCommand $nTransactionID $sParameters";
        if($bIncrementTransactionID) {
                $self->{"$sTransIDType"}++;
        }
        return  $sCMD;
}

sub _genSBTransCmd {
        my ($self, $sCommand, $sParameters, $nTransactionID) = @_;
        return  $self->_genTransCmd($sCommand, $sParameters,$nTransactionID,'SB_TRANS_ID');
}


# [add some low level socket manipulation] #
sub _send {
        my ($self, $sCommand, $bDontAddNewLine) = @_;

        if(!$bDontAddNewLine) {
                $sCommand.="\r\n";
        }
        send($self->{SOCKET},$sCommand,0) || $self->_debug("Error sending packets ($sCommand)");
        $self->{nLastPing} = time();
        $self->_debug("_send()::>> $sCommand");
}

sub _requestSBSession {
        my ($self, $sUserID) =@_;
        my $buf = "";

        # XFR 15 SB
        $self->_send($self->_genTransCmd("XFR","SB"));

        while( 1 ) {
                $buf = $self->_recv();
                if(!defined($buf)) {
                        next;
                }
                if($buf =~/^XFR /) {
                        last;
                }
                CORE::sleep(1);
        }
        $self->_initiateSwitchboardServerConnection($sUserID, $buf);
}

sub _sendToSB {
        my ($self, $sUserID, $sCommand, $bDontAddNewLine) = @_;

        if(!defined($hChatSess{$sUserID}->{'SB_SOCKET'}) ) {
                #return undef;
                # -- connect to switchboard server and send the message
                $self->_requestSBSession($sUserID);
        }

        if(!$bDontAddNewLine) {
                $sCommand.="\r\n";
        }

        send($hChatSess{$sUserID}->{'SB_SOCKET'},$sCommand,0) || $self->_debug("Error sending packets ($sCommand)");
        $self->_debug("_sendSB($sUserID)::>> $sCommand");
}


sub _recvFromSB {
        my ($self, $sUserID, $bDontRemoveEOL, $nTimeOut) = @_;
        my $sCommand = "";

        if(!defined($hChatSess{$sUserID}->{'SB_SOCKET'})) {
                return undef;
        }

        if(!$self->isConnected($hChatSess{$sUserID}->{'SB_SOCKET'})) {
                close($hChatSess{$sUserID}->{'SB_SOCKET'});
                undef $hChatSess{$sUserID}->{'SB_SOCKET'};
                return undef;
        }

        recv($hChatSess{$sUserID}->{'SB_SOCKET'},$sCommand,512, 0);
        if(!length($sCommand)) {
                # $self->_debug("_recvSB(): No content");
                return undef;
        }

        if(!$bDontRemoveEOL) {
                $sCommand =~s/\r\n$//;
        }
        $self->_debug("_recvFromSB($sUserID): << $sCommand");
        if($sCommand eq "") {
                return $sCommand;
        }
        else {
                return $self->_processSBCommand($sUserID,$sCommand);
        }
}

sub _processSBCommand {
        my ($self, $sUserID, $sCommand) = @_;

        #ToDo:: add code to process more than one command at once
        if($sCommand!~/^MSG / && $sCommand=~/\r\n/ ) {
                $self->_processSBCommand($sUserID, substr($sCommand,index($sCommand,"\r\n")+2) );
                $sCommand = substr($sCommand,0, index($sCommand,"\r\n"));
        }

        $self->_debug("_processSBCommand(): $sUserID, $sCommand");
        my @bit = split(/ /,$sCommand);

        if($bit[0] eq "JOI") {
                $hChatSess{$sUserID}->{'SB_READY'}=1;
                # send message from Queue waiting for $sUserID
                $self->_debug("$sUserID has joined the chat");
                foreach (@{ $hChatSess{$sUserID}->{'InQueue'} }) {
                        $self->_sendSBMessage($sUserID, $_);
                }
        }
        elsif($bit[0] eq "MSG") {
                $self->_parseSBMessage($sUserID, $sCommand);
                #$self->_recvFromSB($sUserID,$bDontRemoveEOL);
        }
        elsif($bit[0] eq "BYE") {
                $self->_debug("Closing connection to Switchboard server ($sUserID) ");

                close($hChatSess{$sUserID}->{'SB_SOCKET'});
                $hChatSess{$sUserID} = undef;
        }
        return $sCommand;
}

sub _parseSBMessage {
        my ($self, $sUserID, $sCommand) =@_;

        if($sCommand eq "") {
                return undef;
        }

        # MSG slaff2@hotmail.com slaff 124
        my @msgBits = split(/ /, $sCommand);
        my $msgLength =(split(/\r\n/,$msgBits[3]))[0];

#warning:: length is not always this bit
        #Example: 91\r\nMIME-Version:

        my $nStartPos = index($sCommand,"\r\n");

        if($nStartPos < 1) {
                # This message is not a normal chat message
                return undef;
        }

        my $sMSNMessage = substr($sCommand,$nStartPos, $msgLength+2);
	
	if($sMSNMessage=~/Content-Type: text\/x-msmsgscontrol/g) {		 
                if($sMSNMessage =~m/TypingUser: (.*?)\r\n/) {                        			
                	$self->_addEventToOutQueue("utf8", $1, 'SYS', 'TYPING', "");
                }
        }
	else {
                my $sHeader = substr($sMSNMessage, 0, index($sMSNMessage,"\r\n\r\n"));
                my $sContent = substr($sMSNMessage, index($sMSNMessage,"\r\n\r\n")+4);
#warning:: experimental stuff
                from_to($sContent, "utf8","cp1251");
#warning:: end experimental section
                #$self->_addEventToOutQueue("utf8", $sUserID, 'USR', 'NEW_MSG', $sContent);
                $self->_addEventToOutQueue("cp1251", $sUserID, 'USR', 'NEW_MSG', $sContent);
        }

        if( ($nStartPos+$msgLength)<length($sCommand)) {
                $self->_processSBCommand($sUserID, substr($sCommand, $nStartPos+$msgLength+2));
        }
}

sub _recvAllFromSB {
        my $self = shift;
        foreach my $sUserID (keys(%hChatSess)) {
                $self->_recvFromSB($sUserID);
        }
}

sub _recv {
        my ($self, $bDontRemoveEOL, $nTimeOut) = @_;
        my $sCommand = "";

        recv($self->{SOCKET},$sCommand,512, 0); # || return undef;

        if(!length($sCommand)) {
                return undef;
        }

        if($sCommand!~/\r\n$/ && $sCommand=~/\r\n/) {
                $self->_debug("_recv:: May be this is not complete command ($sCommand)");
                $sCommand .= $self->_recv();
        }

        if(!$bDontRemoveEOL) {
                $sCommand =~s/\r\n$//;
        }
        $self->_debug("_recv():: << $sCommand");

        $self->_processNSCommand($sCommand);

        return $sCommand;
}

sub _processNSCommand {
        my ($self, $sCommand) = @_;

        if(!length($sCommand)) {
                return undef;
        }

        if($sCommand!~/^MSG / && $sCommand=~/\r\n/ ) {
                $self->_processNSCommand(substr($sCommand,index($sCommand,"\r\n")+2) );
                $sCommand = substr($sCommand,0, index($sCommand,"\r\n"));
        }

        #$self->_debug("_processNSCommand(sCommand): $sCommand");

        my $sMSMCommand = (split(/ /, $sCommand))[0];
        $sMSMCommand =~ s/\r\n$//;  # remove \r\n

        if($sMSMCommand=~/\d\d\d/) {
                $self->error($sMSMCommand);
        }
        elsif($sMSMCommand eq "MSG") {
                $self->_processMSG($sCommand);
        }
        elsif(($sMSMCommand eq "CHL")) {
                $self->_processCHL($sCommand);
        }
        elsif(($sMSMCommand eq "LST")) {
                $self->_processLST($sCommand);
        }
        elsif(($sMSMCommand eq "LSG")) {
                $self->_processLSG($sCommand);
        }
        elsif(($sMSMCommand eq "NLN")) {
                $self->_processNLN($sCommand);
        }
        elsif(($sMSMCommand eq "FLN")) {
                $self->_processFLN($sCommand);
        }
        elsif(($sMSMCommand eq "ILN")) {
                $self->_processILN($sCommand);
        }
        elsif(($sMSMCommand eq "OUT")) {
                $self->_processOUT($sCommand);
        }
        elsif(($sMSMCommand eq "RNG")) {
                $self->_processRNG($sCommand);
        }
        elsif(($sMSMCommand eq "QRY")) {
                $self->_debug("Server ping is accepted");
        }
}

sub _processMSG {
        my ($self, $sCommand) = @_;

        #MSG Hotmail Hotmail 414
        my $msgLength = (split(/ /,$sCommand))[3];  # X\r\nMIME ...
        $msgLength =(split(/\r\n/,$msgLength))[0]; # integer(X)


        my $nStartPos = index($sCommand,"\r\n");

        if($nStartPos < 1) {
                # This message is not a normal chat message
                $self->_recv();
                return undef;
        }

        my $sMSNMessage = substr($sCommand,$nStartPos, $msgLength+2);
        $self->_debug("_processMSG(): Server message: $sMSNMessage");

        # proccess according the content type
        $self->_debug( "_processMSG(): Message: [$sMSNMessage]\n");


        if($sMSNMessage=~/Content-Type: text\/x-msmsgsinitialemailnotification;/g ) {
                # Content-Type: text/x-msmsgsinitialemailnotification; -- if we have emails in our email
                #       Inbox-Unread: 4
                #       Folders-Unread: 0

                #ToDo:: add initial message notification
                $self->_debug("_processMSG(): we have new MSN emails");
                $sMSNMessage=~s/\r//g;
                if($sMSNMessage=~/Inbox\-Unread\: (.*?)\n/g) {
                        $self->_debug("We have $1 unread messages");
                        $self->{UNREAD_EMAILS} = $1;
                }
                $self->_addEventToOutQueue("utf8", -1, 'SYS', 'NEW_EMAIL', $self->{UNREAD_EMAILS});
        }
        elsif($sMSNMessage=~/Content-Type: text\/x-msmsgsemailnotification;/g) {
                $self->_parseEmailNotification($sMSNMessage);
        }
        elsif($sMSNMessage=~/Content-Type: text\/x-msmsgsprofile;/g) {
                if($sMSNMessage=~/MSPAuth\:  (.*?)/g) {
                        $self->_debug("Authantication string is: $1");
                        $self->{AUTH_STRING} = $1;
                }
        }       

        if( ($nStartPos+$msgLength)<length($sCommand)) {
                $self->_processNSCommand(substr($sCommand, $nStartPos+$msgLength+2));
        }
}

sub _processLSG {
        my ($self, $sCommand) = @_;
        my @cmdPart = split(/ /,$sCommand);
        #Old: LSG 5 21 2 4 1 Unknown 0
        #New: LSG 0 Other%20Contacts 0
        my ($nGroupID, $sGroupName) =( $cmdPart[1], $cmdPart[2]);
        $sGroupName =~ tr/+/ /;
        $sGroupName =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
#warning:: experimental code
        from_to($sGroupName,"utf8","cp1251");
        ${$self->{_hGroups}}{$nGroupID}=$sGroupName;
        $self->_debug("_processLSG(): $sGroupName");

        $self->{_nGroupsToFetch}--;
}
sub _processLST {
        my ($self, $sCommand) = @_;
        my @cmdPart = split(/ /,$sCommand);

        $self->_debug("Processing LST command ($sCommand)");

        #Old: LST 5 FL 21 3 3 deyanp@hotmail.com Deyan 0
        #New: LST user3@passport.com user3 11 0,1,2,3
        if($cmdPart[3] & 1) { # ex: Forward List
                my ($sUserID, $sUserName, $sUserGroupList) = ($cmdPart[1],$cmdPart[2],$cmdPart[4]);
                ${$self->{_hUsers}}{$sUserID}->{'NAME'}=$sUserName;
                ${$self->{_hUsers}}{$sUserID}->{'GROUP_ID'}=$sUserGroupList;
                $self->_debug("_processLST(): $sUserName ( $sUserID ) -> ($sUserGroupList) ");
        }
        $self->{_nUsersToFetch}--;
}


sub _processCHL {
        my ($self, $sCommand) = @_;
        my ($sCMD, $nTransID,$sParam) = split(/ /,$sCommand);
        $sParam =~s/\r\n//; # remove ending \r\n if any
        $self->_send($self->_genTransCmd("QRY","msmsgs\@msnmsgr.com 32\r\n".
                                                                        md5_hex($sParam.'Q1P7W2E4J9R8U3S5')   ),1);
        $self->_recv();
}

sub _processOUT {
        my ($self, $sCommand) = @_;
        my ($sCMD, $sParam) = split(/ /,$sCommand);

        if($sParam eq "OTH") {
                $self->_error("Logged from another client");
                $self->_addEventToOutQueue("utf8", -1, 'SYS', 'DISCONNECT', 'Logged from another client');
                $self->disconnect();
        }
        elsif($sParam eq "SSD") {
                $self->_error("Server is shutting down for maintanace");
                $self->_addEventToOutQueue("utf8", -1, 'SYS', 'DISCONNECT', 'Server is shutting down for maintanace');
                $self->disconnect();
        }
        $self->_addEventToOutQueue("utf8", -1, 'SYS', 'DISCONNECT', 'Server disconnect');
}

sub _processNLN {
        my ($self, $sCommand) = @_;
        # NLN NLN deyanp@hotmail.com Deyan
        my ($sMSNCmd, $sStatus, $sUserID, $sUserName) = split(/ /,$sCommand);
        $self->_debug("$sUserName ($sUserID) changes state to $sStatus");

        if($sStatus eq "NLN") {
                $self->_addEventToOutQueue("utf8", $sUserID, 'SYS', 'STAT', 'ONL');
        }
        elsif($sStatus eq "BSY") {
                $self->_addEventToOutQueue("utf8", $sUserID, 'SYS', 'STAT', 'BSY');
        }
        elsif(($sStatus eq "IDL") || ($sStatus eq "AWY") || ($sStatus eq "PHN") || ($sStatus eq "LUN")) {
                $self->_addEventToOutQueue("utf8", $sUserID, 'SYS', 'STAT', 'IDL');
        }
        $self->_recv();
}

sub _processFLN {
        my ($self, $sCommand) = @_;
        # FLN deyanp@hotmail.com
        my ($sMSNCmd, $sUserID) = split(/ /,$sCommand);
        $self->_debug("$sUserID goes offline ");
        $self->_addEventToOutQueue("utf8", $sUserID, 'SYS', 'STAT', 'OFF');
        $self->_recv();
}

sub _processILN {
         my ($self, $sCommand) = @_;
        # ILN 5 BSY mircho@hotmail.com momche
        my ($sStatus, $sUserID) = (split(/ /,$sCommand))[2,3];
        $self->_debug("$sUserID changes status to  $sStatus");
=head Sub statuses
    *  BSY - Busy
    * IDL - Idle
    * BRB - Be Right Back
    * AWY - Away
    * PHN - On the Phone
    * LUN - Out to Lunch
=cut
        if($sStatus eq "FLN") {
                $self->_addEventToOutQueue("utf8", $sUserID, 'SYS', 'STAT', 'OFF');
        }
        elsif($sStatus eq "BSY") {
                $self->_addEventToOutQueue("utf8", $sUserID, 'SYS', 'STAT', 'BSY');
        }
        elsif(($sStatus eq "IDL") || ($sStatus eq "AWY") || ($sStatus eq "PHN") || ($sStatus eq "LUN")) {
                $self->_addEventToOutQueue("utf8", $sUserID, 'SYS', 'STAT', 'IDL');
        }
        elsif($sStatus eq "NLN") {
                $self->_addEventToOutQueue("utf8", $sUserID, 'SYS', 'STAT', 'ONL');
        }

        $self->_recv();
}

sub _processRNG {
         my ($self, $sCommand) = @_;
        # RNG 312008 207.46.108.74:1863 CKI 1053170326.21942 galfy@linux-bg.hypermart.net galfy@linux-bg.hypermart.net%20(E-mail%20Address%20Not%20Verified)
        my @bit = split(/ /,$sCommand);
        $self->_debug($bit[5]." wants to send message (Switchbord server is ".$bit[2].", MD5 key is: $bit[4])");
        $self->_connectToSwitchboardServer(@bit);
        $self->_recv();
}

sub _urlencode {
    my $self = shift();
    my $ask = shift();

    my @a2 = unpack("C*", $ask);
    my $s2 = "";
    while (@a2) {
        $s2 .= sprintf "%%%X", shift @a2;
    }
    $self->_debug("_urlencode(): [$ask] -> [$s2] ");
    return $s2;
}

sub _parseEmailNotification {
        my ($self, $sCommand) =@_;
        my $sFrom = "";
        my $sSubject = "";
        my ($i, $max) = ();

=head Sample eMail Notification
MSG Hotmail Hotmail 424
MIME-Version: 1.0
Content-Type: text/x-msmsgsemailnotification; charset=UTF-8

From: SourceForge.net
Message-URL: /cgi-bin/getmsg?msg=MSG1055924463.70&start=1538193&len=2197&curmbox=ACTIVE
Post-URL: https://loginnet.passport.com/ppsecure/md5auth.srf?lc=1033
Subject: =?"ISO-8859-1"?Q?[dosbox_-_General_Discussion]_RE:_Please_compile_the_CVS_for_Win32?=
Dest-Folder: ACTIVE
From-Addr: noreply@sourceforge.net
id: 2
=cut

        $sCommand =~s/\r//g;
        my @aLines = split(/\n/,$sCommand);
        for($i=0, $max=@aLines;$i<$max; $i++) {
                my ($var, $val)=split(/: /, $aLines[$i]);
                if($var eq "From") {
                        $sFrom = $val;
                }
                elsif($var eq "From-Addr") {
                        $sFrom =$sFrom."( $val )";
                }
                elsif($var eq "Subject") {
                        if($val=~/Q\?(.*?)\?=$/) {
                                $sSubject = $1;
                        }
                        else {
                                $sSubject = $val;
                        }
                }
        }
        $self->_debug("New email from: $sFrom; Subject:  $sSubject");
        $self->_addEventToOutQueue("utf8", $sFrom, 'SYS', 'NEW_EMAIL', $sSubject);
}

sub _makeSocketNonBlocking {
        my ($self, $fdSocket) = shift;

        if(!defined $fdSocket) {
                $fdSocket = $self->{SOCKET};
        }

        if(defined($fdSocket)) {
                # -- make it non-blocking socket
                if(!defined $fdSocket->blocking(0)) {
                        $self->_debug("Error with setting blocking: $!");
                }
                else {
                        $self->_debug("Socket is made non-blocking");
                        $self->{SOCKET_IS_NONBLOCKING}=1;
                }
        }
}

1;

__END__