package IM::YIM;
use base 'IM::Base';
#######################################
#	Class for Yahoo Messenger protocol
#	Copyright: Slavej Karadjov -2003
#######################################
use strict;
use Data::Dumper; # for debug purposes
$|=1;

my $nomodule = 0;
eval ("require '/usr/local/apache/im/proxy/Net/YahooMessenger.pm'"); # for debug purposes
if($@) {
        $nomodule++;
        $@ = '';
}
if($nomodule) {
        eval ("use Net::YahooMessenger");
        if($@) {
                print "Can't find Net::YahooMessenger\n";
                exit;
        }
}

my $nLastMessageTo = 0; # values are UIN

my $oYIM = {};

# [Constructor] #
sub new {
        my $proto = shift;
	my $class = ref($proto) || $proto;
	my $self = $class->SUPER::new(@_);
	$self->{nPingTimeOut} = 120;
	$self->{CONTACT_LIST_READY} = 0;

	$self->{PROTOCOL} = "YIM";
        $self->{SCREEN_NAME} = undef;
        $self->{SOCKET}  = undef;
        $self->{UNREAD_EMAILS}  = undef;
        $self->{AUTH_STRING} = undef; # yahoo cookie in fact


        # ...     YIM  

        bless( $self );

        return $self;
}

# :: public functions
sub connect {
        my $self = shift();
        $self->{CLIENT_IP}= shift(); # IP of the client
        $self->_debug("Connecting to YIM  server");

#warning:: arhum ...
        while(1) {
                $oYIM = Net::YahooMessenger->new(
                        id                => $self->{USER_ID},
                        password          =>$self->{USER_CREDENTIALS},
                        #hostname      => 'scs.yahoo.com'
                        hostname      => 'scs.msg.yahoo.com'
                );
                $oYIM->set_event_handler(bless( {
						  'oIM'=> $self
						}, 'CommandLineEventHandler'));

                $oYIM->login() && last;

                $self->_debug("connect(): Invalid Login. Make sure your ID and PASSWORD are entered correctly.");
                return 0;
        }
        $self->_debug("connect(): Almost ready");
        return 1;
}


sub error {
        my $self=shift;
        my $sText = shift;

        #my %ERROR_CODES = (
        #               911=>"Authentication failed", #   -      YIM
        #       );
        # ...
        #$self->_debug("::YIM Error::".$sText."->".$ERROR_CODES{$sText});
        $self->_debug("::YIM Error:: $sText");
        exit();
}

sub addUser {
        my ($self, $sID, $sName, $sGroupList) = @_;
        ${$self->{_hUsers}}{$sID}->{'NAME'} = $sName;
        ${$self->{_hUsers}}{$sID}->{'GROUP_ID'} = $sGroupList;
}

sub addUserStatus {
        my ($self, $sID, $sStatus) = @_;
        ${$self->{_hUsers}}{$sID}->{'STATUS'} = $sStatus;
}

sub addGroup {
        my ($self,$sName) = @_;
        if(!exists(${$self->{_hGroups}}{$sName})) {
                ${$self->{_hGroups}}{$sName} = $sName;
        }
}

sub getContactList {
        my($self) = shift();

        $self->_debug("getContactList(): Retreiving Contact List");

        # get all users and user groups from remote contact list
        while(!$self->{CONTACT_LIST_READY}) {
                print "+";
                $oYIM->do_one_loop();
        }
        $self->_debug("getContactList(): Start generating contact list");
        return $self->_generateContactListString();
}

sub pingIMServer {
        my $self = shift();
        #  PING  
        $self->_debug("pingIMServer(): do ping");
        $oYIM->ping();
        $self->{nLastPing} = time();
}

sub changeState {
        my($self, $sStatusCode) = @_;

        #Yahoo Statuses :
        #       0 - I'm Available
        #       1 - Busy
        #       2 - Sleep

        my %STATUS = (
                'ONLINE' => 0,
                'OCCUPIED' => 1,
                'AWAY' => 2,
		'HIDDEN' => 12
                );
        #        .       ONLINE, OFFLINE  HIDDEN
        $oYIM->change_state($STATUS{$sStatusCode}, " ");

        if($sStatusCode eq "ONLINE") {
                # if we are online
                foreach(keys(%{$self->{_hUsers}})) {
                        if(${$self->{_hUsers}}{$_}->{'STATUS'} ) {
                                $self->_addEventToOutQueue('cp1251', $_, "SYS", "STAT", 'ONL');
                        }
                        else {
                                $self->_addEventToOutQueue('cp1251', $_, "SYS", "STAT", 'OFF');
                        }
                }
        }
        $self->_debug("Status Changed to: $sStatusCode");
}

sub _getProtocolStatus{
	my ($self, $sStatus) = @_;
	my $sNewStatus = undef;
	if($sStatus eq "ONL") {
		$sNewStatus = "ONLINE";
	}
	elsif($sStatus eq "HDN") {
		$sNewStatus = "HIDDEN";
	}
	elsif($sStatus eq "IDL") {
		$sNewStatus = "AWAY";
	}
	elsif($sStatus eq "BSY") {
		$sNewStatus = "OCCUPIED";
	}
	return $sNewStatus;
}

sub sleep {
        CORE::sleep(2);
        #         ,
        #         ,     
}

sub getIncomingEvents {
        my $self = shift;
        #      YIM 
        $oYIM->do_one_loop();
}

sub hasIncomingEvents {
        return 0;
}

sub sendTextMessage {
	my($self, $to, $text) = @_;
	$self->_sendTextMessage($to, $text);
}

sub disconnect {
        my($self) = shift();
        $self->_debug("Disconnecting From YIM Server");
        if($self->isConnected()) {
                #ToDo:   
        }
        $self->{_CONNECTED} = 0;
        #exit();
}


# ::private functions

sub _encodeMessage {
        my ($self, $sMessage) =@_;
        #ToDo: do some charset etc. encoding
        return $sMessage;
}

sub _sendTextMessage {
        my ($self, $toYID, $sMessage) = @_;
        $nLastMessageTo = $toYID;
        $oYIM->send($toYID, $sMessage);
        $self->_debug("_sendTextMessage(".${$self->{_hUsers}}{$toYID}->{'NAME'}." [$toYID] ): $sMessage");
        $self->{nLastPing} = time();
}

package CommandLineEventHandler;
use base 'Net::YahooMessenger::EventHandler';
use strict;
use Data::Dumper; # for debug purposes

use constant STATUS_MESSAGE => [
        "I'm Available",
        'Be Right Back',
        'Busy',
        'Not At Home',
        'Not At My Desk',
        'Not In The Office',
        'On The Phone',
        'On Vacation',
        'Out To Lunch',
        'Stepped Out',
];


sub UnImplementEvent
{
        my $self = shift;
        my $event = shift;
        #$self->_debug("UnImplementEvent(): ".Dumper($event));
        $self->_debug("UnImplementEvent(): $event");
}


sub Login
{
	my $self = shift;
        my $event = shift;

        my $yahoo = $event->get_connection;
        $self->_debug("Logged to Yahoo IM Server");

        $self->{oIM}->{_CONNECTED} = 1;

        my @buddy_name = map {
                $_->name()
        }  $yahoo->buddy_list;

        my @buddy_status = map {
                $_->online()
        }  $yahoo->buddy_list;

        my @buddy_group = map {
                $_->group()
        } $yahoo->buddy_list;

        for(my $i=0; $i<@buddy_name; $i++) {
                $self->{oIM}->addUser($buddy_name[$i],$buddy_name[$i],$buddy_group[$i]);
                $self->{oIM}->addUserStatus($buddy_name[$i],$buddy_status[$i]);
                $self->{oIM}->addGroup($buddy_group[$i]);
        }
	$self->{oIM}->{CONTACT_LIST_READY} = 1;
}

sub InvalidLogin
{
        my $self = shift;
        my $event = shift;

        $self->_debug("Unable To Log To Yahoo IM Server :(");
	$self->{oIM}->{_CONNECTED}=0;
}


sub GoesOnline
{
        my $self = shift;
        my $event = shift;

        $self->_debug( $event->from." goes online");
        $self->{oIM}->_addEventToOutQueue('cp1251', $event->from, 'SYS', 'STAT', 'ONL');
}


sub GoesOffline
{
        my $self = shift;
        my $event = shift;

        if ($event->from) {
                $self->_debug( $event->from." goes offline");
                $self->{oIM}->_addEventToOutQueue('cp1251', $event->from, "SYS", "STAT", 'OFF');
        } else {
                $self->_debug("You have been logged off as you have logged in on a different machine");
                $self->{oIM}->{_CONNECTED}=0;
        }
}


sub ChangeState
{
        my $self = shift;
        my $event = shift;

        my $busy_status = $event->busy == 1 ?
                        '(Busy) ' :
                        $event->busy == 2 ?
                                '(Sleep) ' : '';

        my $message;
        if ($event->status_code == 99) {
                $message = sprintf "[%s] %sTransit to '%s'\n",
                        $event->from, $busy_status, $event->body;
        }
        else {
                $message = sprintf "[%s] %sTransit to '%s'\n",
                        $event->from, $busy_status, STATUS_MESSAGE->[$event->status_code];
=head
"I'm Available", 0
        'Be Right Back', 1
        'Busy', 2
        >3 Away
=cut
                if($event->status_code==0) {
                       $self->{oIM}->_addEventToOutQueue('cp1251', $event->from, "SYS", "STAT", 'ONL');
                }
                elsif($event->status_code==2) {
                        $self->{oIM}->_addEventToOutQueue('cp1251', $event->from, "SYS", "STAT", 'BSY');
                }
                else {
                       $self->{oIM}->_addEventToOutQueue('cp1251', $event->from, "SYS", "STAT", 'IDL');
                }

        }
        print $message;
}


sub NewFriendAlert
{
        my $self = shift;
        my $event = shift;

        my $message = sprintf "[system] New Friend Alert: %s added %s as a Friend.\n",
                $event->from, $event->to;
        $message .= sprintf "and also sent the following message: %s\n",
                $event->body;
        print $message;
}

sub EmailEvent {
        my $self = shift;
        my $event = shift;

        $self->_debug("EmailEvent(): ********************************************\n");

        my $count = $event->count();
        my $from_who = $event->from();
        my $email = $event->email();
        my $subj = $event->subject();

        $self->_debug("EmailEvent(): From: ".$event->from().",Email:".$event->email().",Count:".$event->count()."\n");

        if($from_who && $email && $subj) {
                my $sFrom = "$from_who <$email>";
                $self->{oIM}->_addEventToOutQueue("utf8", $sFrom, 'SYS', 'NEW_EMAIL', $subj);
        }
        elsif($count>0) {
                $self->{oIM}->_addEventToOutQueue("utf8", -1, 'SYS', 'NEW_EMAIL', $count);
        }
}


sub ReceiveMessage
{
        my $self = shift;
        my $event = shift;

        my $body = $event->body;
        $body =~ s{</?(?:font|FACE).+?>}{}g;
        my $message = sprintf "[%s] %s\n", $event->from, $body;
        $self->_debug("new message: $message");
        $self->{oIM}->_addEventToOutQueue('cp1251', $event->from, "USR", "NEW_MSG", $body);
}

sub ServerIsAlive {
        my $self = shift;
        my $event = shift;

        $self->_debug("ServerIsAlive(): Yep");
}

sub NullEvent {
        my $self = shift;
        my $event = shift;
#ToDo:: We should catch UserIsTyping messages somehow
        # 'source' => '4|t_sivenova|49|TYPING|1|t_sivenova|14| |13|1|5|slaff2|1002|1|'
        #                      4|t_sivenova|49|TYPING|1|t_sivenova|14| |13|0|5|slaff2|1002|1|
#ToDo:: may be - message from user not in the list or add user to list
        #5|galfyo|10|99|19|-1615258025|
        if(defined($event->source)) {
                my @bit = split (/\xC0\x80/, $event->source);
                # I'm not sure it the 0 bit is containig the command
                if($bit[0]==4 && $bit[3] eq "TYPING") {
                        $self->_debug("NullEvent(): User ".$bit[1]." is typing");
                }
        }
        $self->_debug("NullEvent(event->code): ".Dumper($event->source));
}

sub _debug {
        my($self) = shift();
        my $sText  = shift();
	$self->{'oIM'}->_debug($sText);
}

1;
__END__
