#!/usr/bin/perl -w
use Getopt::Std;
getopts('l:', \%opts);

use FindBin;
use lib ("$FindBin::Bin/.", "$FindBin::Bin/extra"); #speed up loading my libraries by putting "." path in first place
use Proc::Daemon; # work as Unix daemon
if($opts{'l'}) {
	Proc::Daemon::setLogFile($opts{'l'});
}
Proc::Daemon::Init();

use Socket;

# [Configuration settings] #
$|=1; # autoflush buffers
my $EOL = "\015\012";
my $sPort = 10001;
my $sProtocol = getprotobyname('tcp');
$sProtocol= $1 if $sProtocol =~ /(\d+)/; # untaint port number

# [Load Gateway Host Modules] #
#use GW::BGDATE;
#use GW::LINUXBG;

# [Load IM Protocol Modules] #
my %imProtocols = (
		"MSN" =>1,
		"ICQ" =>1,
		"YIM" =>1,
		"AIM" =>1,
		"LIM" =>1, # LAFFER instant messenger
		# "Gadu" => 1, # Gadu-Gadu protocol
	);

# [Functions] #
sub logMsg {
    	print "IM_PROXY: @_ at ", scalar localtime, "\n";
}

# [Signal Handling] #
sub INT_handler {
    # close all files.
    logMsg("caught signal INT. exiting");
    $oGW->brokenConnection();
    $oGW->shutdown();
    exit(0);
}

sub TERM_handler {
    # close all files.
    logMsg("caught signal TERM. exiting");
    $oGW->brokenConnection();
    $oGW->shutdown();
    exit(0);
}

sub KILL_handler {
    # close all files.
    logMsg("caught signal KILL. exiting");
    $oGW->brokenConnection();
    $oGW->shutdown();
    exit(0);
}

$SIG{'INT'} = 'INT_handler';
$SIG{'TERM'} = 'TERM_handler';
$SIG{'KILL'} = 'KILL_handler';
$SIG{'CHLD'} = 'IGNORE';


socket(SERVER_HANDELER, PF_INET, SOCK_STREAM, $sProtocol)        || die "socket: $!";
setsockopt(SERVER_HANDELER, SOL_SOCKET, SO_REUSEADDR, pack("l", 1))   || die "setsockopt: $!";
bind(SERVER_HANDELER, sockaddr_in($sPort, INADDR_ANY))        || die "bind: $!";
listen(SERVER_HANDELER,SOMAXCONN)    || die "listen: $!";

# [Starting Server] #
logMsg("server started on port $sPort");

for(
     $kidpid = 0;
     (my $paddr = accept(CLIENT_HANDLER,SERVER_HANDELER) ) ||  $kidpid;
     $kidpid=0, close(CLIENT_HANDLER)
     )  {
	next if !$paddr;
        my($port,$iaddr) = sockaddr_in($paddr);
        my $name = gethostbyaddr($iaddr,AF_INET);

        logMsg("connection from $name [".inet_ntoa($iaddr). "] at port $port");

	if(!defined($kidpid = fork())) {
		# fork returned undef, so failed
   		 logMsg("unable to fork new process");
  	}
	elsif($kidpid==0) { # fork returned 0, so this branch is child
		my $sIMProtocol = "";
		my $sIMUserID = "";
		my $sIMUserCredential = "";
		my $sIMLocalUserID = "";

		# Set Timeout for sending initial connection string
		$SIG{'ALRM'} = sub {
				alarm(0);
				logMsg("No connection string within 5 seconds. Exiting child process");
				exit(0);
				   };
		alarm(5); # 5 wait only 5 seconds for initial connection string

		logMsg("Start handling new process");

		my $buf =<CLIENT_HANDLER>;
		alarm(0); # stop alarm

		if($buf =~ m/^\!C (.*?)\s+(.*?)\s+(.*?)\s+(.*?)$/) {

			logMsg("Got Connection string");
			$sIMProtocol = $1;
			$sIMUserID = $2;
			$sIMUserCredential = $3;
			$sIMLocalUserID = $4;

			logMsg("Protocol: $1; User: $sIMUserID");
		}
		else {
			logMsg("No Connection string. Closing connection");
			exit(0);
		}

		$sIMUserID =~s/\@/\\\@/g; #be carefull when eval-ing special symbols


		if($imProtocols{$sIMProtocol}) { # if this is valid Protocol identifier
				# load it
				eval("use IM::".$sIMProtocol);
		}

		eval("\$oIM = IM::".$sIMProtocol."->new(\"$sIMUserID\",\"$sIMUserCredential\");");
		if($@) {
			logMsg("Unable to initilize $sIMProtocol protocol. Error was:".$!);
			send(CLIENT_HANDLER,"!E Invalid Protocol",0);
			close(CLIENT_HANDLER);
			exit();
		}

		#notice: the idea is to load this module only in the child processes and keep the parent process as thin
		# as possible
		eval("use GW::LINUXBG");

		# initializa gateway module
		#$oGW=GW::BGDATE->new($sIMLocalUserID, $sIMUserID, $sIMProtocol);
		$oGW=GW::LINUXBG->new($sIMLocalUserID, $sIMUserID, $sIMProtocol);

		if(!$oIM->connect( inet_ntoa($iaddr)) ) {
			logMsg("invalid user credentials");
			send(CLIENT_HANDLER,"!E Invalid Credentials",0);
			exit();
		}

		if(!$oGW->keepAlive()) {
			logMsg("Closing client connection. Start listening for message events from $sIMProtocol server");
			send(CLIENT_HANDLER,"/quit",0);
			close(CLIENT_HANDLER);
		}

		# Set Timeout for initial nagotiation
		$SIG{'ALRM'} = sub {
				alarm(0);
				logMsg("Unable to nagotiate within 2 minutes. Exiting child process");
				exit(0);
				   };
		alarm(120); # wait only 2 minutes for initial nagotiation with remote server

		if(my $sContactList = $oIM->getContactList() ) {
			$oGW->sendContactList($sContactList);
		}

		$oIM->changeState('ONLINE');

		alarm(0); # stop the alarm

		my $inBuffer = "";
		while($oIM->isConnected()) {

			if(!$oGW->isUserConnected()) {
				logMsg("Local user is disconnected. Exiting process");
				$oIM->disconnect();
				last;
			}

			if($oGW->hasOutgoingEvents()) {
				$oIM->sendOutgoingEvents($oGW->getOutgoingEventsFromQueue());
			}
			elsif($oIM->isTimeToPingIMServer) {
				$oIM->pingIMServer();
			}

			if($oIM->hasOutgoingEvents()) {
				$oGW->addEventsToQueue($oIM->getIncomingEventsFromQueue());
			}
			else{
				$oIM->sleep();
				# do not rush too fast if there are no mesages
			}

			#$oIM->changeState();
		}

		logMsg("Connection with $sIMProtocol server is broken. Exiting process");
		$oGW->brokenConnection();
		if($oGW->keepAlive()) {
			# close connection from client to proxy
			send(CLIENT_HANDLER,"/quit",0);
			close(CLIENT_HANDLER);
		}
		exit(0);
	}
	else {
		# fork returned nor 0 nor undef
   		# so this branch is parent
    		# waitpid($kidpid, 0);
	}
}


exit();

END {
	Proc::Daemon::Clean();
}
