#!/usr/bin/perl -w
use Socket;
use Digest::MD5 qw(md5_hex);
use strict;
use DBI;
############################################
#
# IM Login Server; (c) Slavej Karadjov - 2003
#
############################################
use Getopt::Std;
our (%opts);
getopts('l:', \%opts);

use FindBin;
use lib ("$FindBin::Bin/.","$FindBin::Bin/lib"); #speed up loading my libraries by putting "." path in first place
use Proc::Daemon; # work as Unix daemon
if($opts{'l'}) {
	Proc::Daemon::setLogFile($opts{'l'});
}
Proc::Daemon::Init();


# 08/07/2003 Basic version is ready

#ToDo:: Set timeout

use constant _INTERNAL_SERVER_ERROR_ => 100;
use constant _UNKNOWN_CMD_  => 400;
use constant _CMD_NOT_IMPLEMENTED_  => 401;
use constant _INCORRECT_CMD_ORDER_ => 402;

# [load configuration] #
require "conf/config.pm";

# [load user_auth modules] #
use UserAuth::LBG;
my $oUser = UserAuth::LBG->new();
use DS::Test;
my $oDS = new DS::Test->new();

my $dbh = DBI->connect("DBI:mysql:".$Config::DB_DATABASE,$Config::DB_USER,$Config::DB_PASS) || die $DBI::errstr;
$oUser->setDbConnection($dbh);
$oDS->setDbConnection($dbh);


# [My Global Variables] #
my $sCommand = "";
my $sUser = "";
my $sPassw = "";
my $sProtoVersion = "";
my $sClientName = "";
my $sMD5Hash = "";
my $sSessCookie = "";
my $sServer = "";

my $sResponse = "";

my %_VALID_CMD = (
                'VER' => 1, # Version nagotiation
                'CAPS' => 2, # Capabilities nagotiation
                'USER' => 3, # User information
                'PASS' => 4, # Password information
                'GOTO'=> 5, # Redirect to messege server
   );

my $nCmdLevel = 1;
my $nLastCmdLevel = $nCmdLevel;


# [Configuration settings] #
$|=1; # autoflush buffers
my $EOL = "\015\012";
my $sPort = 9999;
my $sProtocol = getprotobyname('tcp');
$sProtocol= $1 if $sProtocol =~ /(\d+)/; # untaint port number

# [Signal Handling] #
sub INT_handler {
    # close all files.
    logMsg("caught signal INT. exiting");
    exit(0);
}

sub TERM_handler {
    # close all files.
    logMsg("caught signal TERM. exiting");
    exit(0);
}

sub WARN_handler {
    my $signal = shift;
    logMsg("__WARN__ caught signal: $signal");
}

sub DIE_handler {
    my $signal = shift;
    logMsg("__DIE__ caught signal: $signal");
}


$SIG{'INT'} = 'INT_handler';
$SIG{'TERM'} = 'TERM_handler';
$SIG{'CHLD'} = 'IGNORE';
$SIG{__WARN__} = 'WARN_handler';
$SIG{__DIE__} = 'DIE_handler';

# [Functions] #
sub logMsg {
        print "IM_FRONT_SERVER: @_ at ", scalar localtime, "\n";
}

sub getRandomMD5Hash {
        my $s_id;

        for (8,4,4,4,12)
        {
                $s_id .= uc sprintf "%x", rand(16) for 1..$_;
                $s_id .= '-' unless $_ == 12;
        }
        return $s_id;
}

sub getUniqueCookie {
#ToDo:: create unique cookie
        return md5_hex(getRandomMD5Hash().time());
}

# [command processors] #
sub execVER {
	my @bit=@_;
	$sProtoVersion = $bit[1];
        $sClientName = $bit[2];
        $sResponse  = "VER OK";
	$nLastCmdLevel = $nCmdLevel;
	$nCmdLevel++;
        return 1;
}

sub execCAPS {
	my @bit=@_;
#warning:: not implemented yet
	$sResponse = "CAPS OK";
	$nLastCmdLevel = $nCmdLevel;
	$nCmdLevel++;
	return 1;
}

sub execUSER {
	my @bit=@_;
	if(($sProtoVersion eq "") && ($sClientName eq "")) {
             return -1;
        }
        elsif(!$bit[1])  {
             return -1;
        }
        else {
             $sUser = $bit[1];
             $sMD5Hash=getRandomMD5Hash();
             $sResponse = "USER $sUser MD5 $sMD5Hash";
	     $nLastCmdLevel = $nCmdLevel;
	     $nCmdLevel++;
             return 1;
        }
}

sub execPASS {
	my @bit=@_;
	if($sUser eq "") {
             return -1;
        }
        elsif(!$bit[1]) {
             return -1;
        }
        else {
             my $sRealPass = $oUser->getUserPass($sUser);
             if(!defined($sRealPass)) {
                 $sResponse = "PASS NOK";
                 return -1;
             }

             if($bit[1] ne md5_hex($sRealPass.$sMD5Hash)) {
                 $sResponse = "PASS NOK";
                 return -1;
             }
             else {
                 $sSessCookie = getUniqueCookie();
                 $sServer = $oDS->getFreeServer();
                 $sResponse = "PASS OK".$EOL.
                 	      "GOTO ".$sServer." ".$sSessCookie;
		 my $sUserID = $oUser->getUserIDFromLogin($sUser);
                 $oDS->createUserSess($sSessCookie, $sUserID, $sUser,
		 			$oUser->getContactList($sUserID),
		 			$sProtoVersion, $sClientName, $sServer);
		 $nLastCmdLevel = $nCmdLevel;
		 $nCmdLevel++;
                 return -1;
             }
       }
}


sub processClientCommand {
        my $sCommandLine = shift;
        my $sCommandToReturn = "";

        $sResponse = "";
	my $nResult = undef;

        if($sCommandLine !~ /$EOL$/) {
                        #ToDo: Do something ....
                        return undef;
        }
        $sCommandLine =~ s/$EOL$//;

        logMsg("processClientCommand(): Processing: $sCommandLine");

        my @bit = split(/ /,$sCommandLine);
        if(!$_VALID_CMD{$bit[0]}) {
                # this command is invalid - so disconnect
                $sResponse = _UNKNOWN_CMD_;
                return -1;
        }
        elsif($_VALID_CMD{$bit[0]} != $nCmdLevel) {
                # this command should wait for its order
                print $_VALID_CMD{$bit[0]} ." <<>>".$nCmdLevel;
                $sResponse = _INCORRECT_CMD_ORDER_;
                return -1;
        }

        eval ("\$nResult = exec".$bit[0]."(\@bit);");
        if(@!) {
                logMsg("exec".$bit[0]."(): ".@!);
                $sResponse = _CMD_NOT_IMPLEMENTED_;
                return  -1;
        }
        return $nResult;
}



# [descructors] #
sub DESTROY {
        $dbh->disconnect();
}

# [Main Logic] #
socket(SERVER_HANDLER, PF_INET, SOCK_STREAM, $sProtocol)        || die "socket: $!";
setsockopt(SERVER_HANDLER, SOL_SOCKET, SO_REUSEADDR, pack("l", 1))   || die "setsockopt: $!";
bind(SERVER_HANDLER, sockaddr_in($sPort, INADDR_ANY))        || die "bind: $!";
listen(SERVER_HANDLER,SOMAXCONN)    || die "listen: $!";

# [Starting Server] #
logMsg("server started on port $sPort");

for(
     my $kidpid = 0;
     (my $paddr = accept(CLIENT_HANDLER,SERVER_HANDLER) ) ||  $kidpid;
     $kidpid=0, close(CLIENT_HANDLER)
     )  {
        next if !$paddr;
        my($port,$iaddr) = sockaddr_in($paddr);
        my $name = gethostbyaddr($iaddr,AF_INET);

        logMsg("connection from $name [".inet_ntoa($iaddr). "] at port $port");

        if(!defined($kidpid = fork())) {
                # fork returned undef, so failed
                 logMsg("unable to fork new process");
        }
        elsif($kidpid==0) {
                # fork returned 0, so this branch is child
                $sCommand = "";
                $sUser = "";
                $sPassw = "";
                $sProtoVersion = "";
                $sClientName = "";
                $sMD5Hash = "";
                $sSessCookie = "";

                my $buf="";
                logMsg("Start handling new process");
                while(1) {
                        recv(CLIENT_HANDLER,$buf,20, 0);

                        my $sRes = processClientCommand($sCommand.$buf);

                        if (!defined($sRes)) {
                                $sCommand.=$buf;
                                next;
                        }
                        elsif($sRes == -1) {
                                if($sResponse)  {
                                        $sResponse.=$EOL;
                                        send(CLIENT_HANDLER,$sResponse, 0);
                                }
                                close(CLIENT_HANDLER);
                                logMsg("exiting. Response: $sResponse");
                                last;
                        }
                        elsif($sRes == 1 ) {
                                if($sResponse)  {
                                        $sResponse.=$EOL;
					logMsg($sResponse);
                                        send(CLIENT_HANDLER,$sResponse, 0) || die $!;
                                }
                                $sCommand= "";
                                next;
                        }
                }
                # ToDo:: ....
                logMsg("<< o >> Closing connection");
                exit();
        }
        else {
                # fork returned nor 0 nor undef
                # so this branch is parent
                # waitpid($kidpid, 0);
        }
}
exit();

__END__
