#!/usr/bin/perl -w
use Socket;
use Fcntl;
use Digest::MD5 qw(md5_hex);
use strict;
############################################
#
# IM Message Server; (c) Slavej Karadjov - 2003
#
############################################
use Getopt::Std;
our (%opts);
getopts('l:', \%opts);

use FindBin;
use lib ("$FindBin::Bin/lib"); #speed up loading my libraries by putting "." path in first place
use Proc::Daemon; # work as Unix daemon
if($opts{'l'}) {
	Proc::Daemon::setLogFile($opts{'l'});
}
Proc::Daemon::Init();

use FS::LocalDB; # use FrontServer via Local DB functionality
use MS::Util;


use constant _INTERNAL_SERVER_ERROR_ => 100;
use constant _INVALID_STATUS_  => 350;
use constant _UNKNOWN_CMD_  => 400;
use constant _CMD_NOT_IMPLEMENTED_  => 401;
use constant _INCORRECT_CMD_ORDER_ => 402;

my %_VALID_CMD = (
                'USER' => 1, # User authenticate
                'GCL' => 2, # Get Contact List
                'CLU' => 2, # Contact List User
                'CLG' => 2, # Contact List Group
                'STAT'=>3, # Set Presence Status
                'PING' => 4, #Keep Connection Alive
                'PONG' =>4, # Pong response
                'ADDU' => 4, # Add User
		'REQ_AUTH' =>4, # Request Authentication
		'GRANT_AUTH'=>4, # Grant authentication
		'REJ_AUTH' =>4, # Reject authentication
		'AUTH_OK' =>4, # Authorization is granted to user X (ALlow Aadding User)
                'ADDG' => 4, # Add Group
		'RENU' => 4, # Rename user
		'RENG' =>4, # Rename group
		'MVU' =>4, # Move user from group X to group Y
                'DELU' => 4, # Remove User
                'DELG' => 4, # Remove Group
                'MSG' => 4, # Message
		'SRCH' => 4, # Search command
                'QUIT'=> 4 # Quit from Server
        );

my %_VALID_STAT = (
                "ONL" =>1, # Online
                "OFF" =>1, # Offline
                "IDL" => 1, # Idle
                "BSY" =>1, # Busy
                "HDN" => 1 # Hidden
        );

# [My Global Variables] #
my $sCommand = "";
my $sUser = "";
my $sUserID = "";
my $sPassw = "";
my $sProtoVersion = "";
my $sClientID = "";
my $sClientName = "";
my $sSessCookie = "";
my $sServer = "";

my $sResponse = "";

my $nCmdLevel = 1;
my $nLastCmdLevel = $nCmdLevel;

my $bGetEventsForClientFromServer = 0;

# [Start Signal Handling] #
sub INT_handler {
    # close all files.
    logMsg("caught signal INT. exiting");
    exit(0);
}

sub TERM_handler {
    # close all files.
    logMsg("caught signal TERM. exiting");
    exit(0);
}

sub WARN_handler {
    my $signal = shift;
    logMsg("__WARN__ caught signal: $signal");
}

sub DIE_handler {
    my $signal = shift;
    logMsg("__DIE__ caught signal: $signal");
}


$SIG{'INT'} = 'INT_handler';
$SIG{'TERM'} = 'TERM_handler';
$SIG{'CHLD'} = 'IGNORE';
$SIG{__WARN__} = 'WARN_handler';
$SIG{__DIE__} = 'DIE_handler';

# [Initialize MS and FS functionality] #
my $oFS = FS::LocalDB->new();
my $oMS = MS::Util->new();
$oFS->addDBConnection(&MS::Util::getDBConnectionRef);


# [ContactList] #
my $sContactListEncoding = "";
my %hContactListUser = ();
my %hContactListGroup = ();


# [Configuration settings] #
$|=1; # autoflush buffers
my $EOL = "\015\012";
my $sPort = 1111;
my $sProtocol = getprotobyname('tcp');
$sProtocol= $1 if $sProtocol =~ /(\d+)/; # untaint port number

# [Functions] #
sub logMsg {
        print "IM_MESSAGE_SERVER: @_ at ", scalar localtime, "\n";
}

sub processClientCommand {
        my $sCommandLine = shift;
        my $sCommandToReturn = "";
        my $nResult = undef;

        $sResponse = "";

        if($sCommandLine !~ /$EOL$/) {
                        #ToDo: Do something ....
                        return undef;
        }
        $sCommandLine =~ s/$EOL$//;

        logMsg("processClientCommand(): Processing: $sCommandLine");

        my @bit = split(/ /,$sCommandLine);

        if(!$_VALID_CMD{$bit[0]}) {
                # this command is invalid - so disconnect
                $sResponse = _UNKNOWN_CMD_;
                return -1;
        }
        elsif(
                (($nCmdLevel < 3) && ($_VALID_CMD{$bit[0]} != $nCmdLevel)) ||
                (($nCmdLevel > 3) && ($_VALID_CMD{$bit[0]} < 3))
        ) {
                # this command should wait for its order
                print $_VALID_CMD{$bit[0]} ." <<>>".$nCmdLevel;
                $sResponse = _INCORRECT_CMD_ORDER_;
                return -1;
        }

        eval ("\$nResult = exec".$bit[0]."(\@bit);");
        if(@!) {
                logMsg("exec".$bit[0]."(): ".@!);
                $sResponse = _CMD_NOT_IMPLEMENTED_;
                return  -1;
        }
        return $nResult;
}

# [ execCommand Functions] #
sub execUSER {
        my @bit = @_;
        my $sContactList = "";

        $sUser = $bit[1];
        $sSessCookie = $bit[2];

        # Get user data:
        # login, user_id, proto, client_id, server, contact_list
        ($sUser, $sUserID, $sProtoVersion, $sClientID, $sServer,  $sContactList) = getInitialUserData($bit[1],$bit[2]);

        if($sUserID) {
                if(createUserIMSess($sSessCookie, $sUserID, $sProtoVersion, $sClientID)) { # Create record in IM_SESS table
                        _processContactList($sContactList);
                        $nCmdLevel = 2;
                        $sResponse = "USER $sUser OK";
                        return 1;
                }
                else {
                        $sResponse = _INTERNAL_SERVER_ERROR_;;
                        return 1;
                }

        }

        $sResponse = "USER $sUser NOK";
        return -1;
}

sub execGCL {
        my @bit = @_;
        my $nItems = 0;
        my $nContactListID = $bit[1];
# < GCL NEW_CONTACT_LIST_ID ITEMS_TO_FOLLOW ENCODING

        $sResponse = "";
        # Send Groups
        foreach(keys(%hContactListGroup) ) {
                $sResponse .= "CLG ".$_." ".$hContactListGroup{$_}.$EOL;
                $nItems++;
        }

        # Send Users
        foreach(keys(%hContactListUser) ) {
                $sResponse .= "CLU ".$_." ".$hContactListUser{$_}->{NAME}." ".$hContactListUser{$_}->{GROUP_LIST}.$EOL;
                $nItems++;
        }

        $sResponse = "GCL ".time()." ".$nItems." ".$sContactListEncoding.$EOL.
                                $sResponse;
        $sResponse =~s/$EOL$//;

        $nCmdLevel = 3;
        return 1;
}

sub execSTAT {
        my @bit = @_;

        if(!$_VALID_STAT{$bit[1]}) {
                $sResponse=_INVALID_STATUS_;
                return -1;
        }
        $sResponse="";

        sendEventToReverseList($sUserID, 'SYS', 'STAT', $bit[1]);
        if(!$bGetEventsForClientFromServer ) {
                $bGetEventsForClientFromServer = 1;

                # get statuses for all users
		my %hFriends = getUsersStatus( keys(%hContactListUser) );
                foreach(keys(%hFriends)) {
                        $sResponse .="STAT ".$_." ".$hFriends{$_}.$EOL;
                }
        }
        $nCmdLevel = 4;

        #ToDo::Later:: Handle (in)Visability
        $sResponse .= "STAT $sUserID ".$bit[1];
        return 1;
}

sub execPING {
        updateUserIMSess($sSessCookie);
        $sResponse = "PONG";
        return 1;
}

sub execADDU {
        my @bit = @_;
#ToDo::get user nickname
	my $sToNick = $bit[1];

	# add user to contact list
	if($oFS->addUser($sUserID,$bit[1],$sToNick,-1)) {
		# send notification to user to be added
		sendEvent('SYS', 'REQ_AUTH', $bit[1], $sUserID, '', time());

		return 1;
	}
	# return error message

}

sub execREQ_AUTH {
	my @bit = @_;
#warning:: not implemented yet
}

sub execGRANT_AUTH {
	my @bit = @_;
	$oFS->authorizeUser($sUserID,$bit[1]);
	sendEvent("SYS", "AUTH_OK", $bit[1], $sUserID, '');
}

sub execREJ_AUTH {
	my @bit = @_;
#warning:: not implemented yet
}

sub execAUTH_OK {
	my @bit = @_;
#warning:: not implemented yet
}

sub execADDG {
	my @bit = @_;
#warning:: not implemented yet
#	> ADDG GROUP_NAME
#	< ADDG GROUP_NAME NEW_GROUP_ID
# if everything is correct or with error message
#	(-)* there is aleady group with such name
	if(my $newGroupID = $oFS->addGroup($sUserID,$bit[1])) {
		# if we added succeessfully the new group
		$sResponse = "ADDG ".$bit[1]." ".$newGroupID;
		return 1;
	}
	else {
		# return error code
#warning:: this is debug reply only
		$sResponse = "ADD GROUP NOK";
		return 1;
	}
}
sub execRENU {
	my @bit = @_;
#warning:: not implemented yet
#	> RENU USER_ID NEW_NICK
#	< RENU USER_ID OK
# or error
#	- no such user_id in your contact list
	if($oFS->updateContactListUser($sUserID,$bit[1],
				     'name'=>$bit[2])
	) {
		# if we renamed succeessfully this user
		$sResponse = "RENU ".$bit[1]." OK";
		return 1;
	}
	else {
		# return error code
#warning:: this is debug reply only
		$sResponse = "RENU ".$bit[1]." NOK";
		return 1;
	}
}

sub execRENG {
        my @bit = @_;
#warning:: not implemented yet
#	> RENG GROUP_ID NEW_GROUP_NAME
#	< RENG GROUP_ID OK
#or error
#	- no such group_id in your contact list
	if($oFS->updateContactListGroup($sUserID,$bit[1],
				     'name'=>$bit[2])
	) {
		# if we renamed succeessfully this user
		$sResponse = "RENG ".$bit[1]." OK";
		return 1;
	}
	else {
		# return error code
#warning:: this is debug reply only
		$sResponse = "RENG ".$bit[1]." NOK";
		return 1;
	}
}

sub execMVU {
        my @bit = @_;
#warning:: not implemented yet
#	> MVU USER_ID FROM_GROUP_ID TO_GROUP_ID
#	< MVU USER_ID OK
#or error
#	- no such user_id in your contact list
#	- no such FROM_GROUP_ID in your contact list
#	- no such TO_GROUP_ID in your contact list
#	(-)* invalid TO_GROUP_ID - if the group id is special/reserved - 0,1,2
	if($oFS->moveContactListUser($sUserID, $bit[1], $bit[2], $bit[3])) {
		$sResponse = "MVU ".$bit[1]." OK";
		return 1;
	}
	else {
		# return error code
#warning:: this is debug reply only
		$sResponse = "MVU ".$bit[1]." NOK";
		return 1;
	}
}

sub execDELU {
        my @bit = @_;
#warning:: not implemented yet
#	> DELU USER_ID
#	1. < DELU USER_ID OK
#	or error message
#	- no such user_id in your contact list
#	- invalid user_id - if the user id is the id of special  user - admin, support etc
	if($oFS->delContactListUser($sUserID,$bit[1]) ) {
		# if we deleted succeessfully this user
		$sResponse = "DELU ".$bit[1]." OK";
		return 1;
	}
	else {
		# return error code
#warning:: this is debug reply only
		$sResponse = "DELU ".$bit[1]." NOK";
		return 1;
	}
}

sub execDELG {
        my @bit = @_;
#warning:: not implemented yet
# 	DELG GROUP_ID
#	1. < DELG GROUP_ID OK
#	or error message
#	- no such group_id
#	- invalid group_id - if the group id is the id of special group - example 0,1,2
	if($oFS->delContactListGroup($sUserID,$bit[1]) ) {
		# if we deleted succeessfully this user
		$sResponse = "DELG ".$bit[1]." OK";
		return 1;
	}
	else {
		# return error code
#warning:: this is debug reply only
		$sResponse = "DELG ".$bit[1]." NOK";
		return 1;
	}
}

#warning:: message command is not working as expected
sub execMSG {
        my @bit = @_;
#  MSG TYPE ENC FROM/TO LENGTH_IN_BITS MESSAGE_CONTENT
#  MSG text utf8 galfy@bgdate 10 1234567890
        my $sType=$bit[1];
        my $sEncoding = $bit[2];
        my $sTo = $bit[3];
        my $nLength = $bit[4];
        my $sMessage = "";

        if($sType eq "text") {
                my $nStartPos = 9+length($sEncoding)+1+length($sTo)+1+length($nLength)+1;
                $sMessage = substr(join(" ",@bit), $nStartPos, $nLength);
                if(!length($sMessage)) {
                        logMsg("execMSG(): emty message??");
                        $sResponse = "MSG NOK";
                }
                else {
                        $sResponse = "MSG OK";
                        sendEvent("USR", "NEW_MSG", $sTo, $sUserID, $sMessage);
                }
        }
        return 1;
}

sub execSRCH {
	my @bit = @_;
#warning:: not implemented yet
#	> SRCH "URL_QUERY"
}

sub execQUIT {
        #Tell eveybody that I am offline now
	sendEventToReverseList($sUserID, 'SYS', 'STAT', 'OFF');
        deleteUserIMSess( $sSessCookie);
	$oFS->delUserSess($sSessCookie);
        $sResponse = "QUIT";
        return -1;
}

# [helper commands] #
sub _urlencode {
    my $ask = shift();

    my @a2 = unpack("C*", $ask);
    my $s2 = "";
    while (@a2) {
        $s2 .= sprintf "%%%X", shift @a2;
    }
    return $s2;
}

sub _processContactList {
        my $sRawContactList = shift();
        my ($tmpGroups, $tmpUsers) = ();

        if($sRawContactList =~/^(.*?);\[(.*?)\];\[(.*?)\]$/) {
                $sContactListEncoding = $1;
                $tmpGroups = $2;
                $tmpUsers = $3;
        }

        # process groups
        my @pair = split(/\|/,$tmpGroups);
        foreach(@pair) {
                my ($sGroupID, $sGroupName)= split(/,/);
                $hContactListGroup{$sGroupID} = _urlencode($sGroupName);
        }

        # process users
        @pair = split(/\|/,$tmpUsers);
        foreach(@pair) {
                if($_=~/^(.*?),(.*?),\{(.*?)\}$/) {
                        $hContactListUser{$1}->{'NAME'} = _urlencode($2);
                        $hContactListUser{$1}->{'GROUP_LIST'} = $3;
                }
        }
        return 1;
}


# [communication commands] #
sub sendEventsToClient {
        my $CLIENT_HANDLER = shift();

        # Get Events From Server for Client
        my @events = getEventsForClient($sUserID);
	foreach(@events) {
        	my($nEventID, $sType, $sCMD, $sFromUser, $sAttr, $sDate) =  split(/,/);
#ToDo v0.3 :: Implement more commands here
                logMsg("sendEventsToClient():GOT EVENT: ($nEventID, $sType, $sCMD, $sFromUser, $sAttr, $sDate) ");
                if($sType eq "SYS") {
                        if($sCMD eq "STAT") {
                                _sendEventToClient($CLIENT_HANDLER, "STAT $sFromUser $sAttr");
                        }
			delEvent($nEventID);
                }
                elsif($sType eq "USR") {
                        if($sCMD eq "NEW_MSG") {
                                _sendEventToClient($CLIENT_HANDLER,"MSG text cp1251 $sFromUser ".length($sAttr)." $sAttr", \$CLIENT_HANDLER);
				updateEvent($nEventID,
						'new' => 'N'
					   );
                        }

                }

        }
}

sub _sendEventToClient {
        my ($CLIENT_HANDLER, $sText) = @_;
        logMsg("_sendEventToClient(): $sText");
        send($CLIENT_HANDLER, $sText.$EOL,0);
}

# [Main Logic] #
socket(SERVER_HANDLER, PF_INET, SOCK_STREAM, $sProtocol)        || die "socket: $!";
setsockopt(SERVER_HANDLER, SOL_SOCKET, SO_REUSEADDR, pack("l", 1))   || die "setsockopt: $!";
bind(SERVER_HANDLER, sockaddr_in($sPort, INADDR_ANY))        || die "bind: $!";
listen(SERVER_HANDLER,SOMAXCONN)    || die "listen: $!";

# [Starting Server] #
logMsg("server started on port $sPort");

for(
     my $kidpid = 0;
     (my $paddr = accept(CLIENT_HANDLER,SERVER_HANDLER) ) ||  $kidpid;
     $kidpid=0, close(CLIENT_HANDLER)
     )  {
        next if !$paddr;
        my($port,$iaddr) = sockaddr_in($paddr);
        my $name = gethostbyaddr($iaddr,AF_INET);

        logMsg("connection from $name [".inet_ntoa($iaddr). "] at port $port");

        if(!defined($kidpid = fork())) {
                # fork returned undef, so failed
                 logMsg("unable to fork new process");
        }
        elsif($kidpid==0) {
                # fork returned 0, so this branch is child
                $sCommand = "";
                $sUser = "";
                $sProtoVersion = "";
                $sClientName = "";
                $sSessCookie = "";

                my $nCmdLevel = 1;

                my $buf="";
                logMsg("Start handling new process");

                logMsg("Make Socket non blocking");
                fcntl(CLIENT_HANDLER, F_SETFL, O_NONBLOCK) or die "fcntl: $!";
                my $sRes = 0;
                while(1) {
                        # [Send Commands from server to client] #
                        if($bGetEventsForClientFromServer) {
                                sendEventsToClient(\*CLIENT_HANDLER);
                        }

                        #if(!
                        recv(CLIENT_HANDLER,$buf,20, 0);
                        # ) {
                        #       # The connection is broken
                        #       $sRes = 2;
                        #       last;
                        #}

                        if(defined($buf)) {
                                # [Process Commands From Client] #
                                $sRes = processClientCommand($sCommand.$buf);

                                #if(defined($sRes)) {
                                #       print "[$sRes]";
                                #}

                                if (!defined($sRes)) {
                                        $sCommand.=$buf;
                                        next;
                                }
                                elsif($sRes == -1) {
                                        if($sResponse)  {
                                                $sResponse.=$EOL;
                                                send(CLIENT_HANDLER,$sResponse, 0);
                                        }
                                        logMsg("exiting. Response: $sResponse");
                                        last;
                                }
                                elsif($sRes == 1 ) {
                                        if($sResponse)  {
                                                $sResponse.=$EOL;
                                                send(CLIENT_HANDLER,$sResponse, 0) || die $!;
                                        }
                                        $sCommand= "";
                                        $buf = "";
                                        next;
                                }
                        }

                }
                # ToDo:: Log Events / Implement Rate Limit
                logMsg("Exited with return value: $sRes");
                if($nCmdLevel>3) {
                        execQUIT();
                }
                logMsg("<< o >> Closing connection");
                close(CLIENT_HANDLER);
                exit();
        }
        else {
                # fork returned nor 0 nor undef
                # so this branch is parent
                # waitpid($kidpid, 0);
        }
}


exit();

__END__
